#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface


class Levy(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        params = [0.55] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        return None

    def _run(self, params):
        params = np.array(params)
        params = 20 * params - 10  # rescale onto [-10, 10]
        w = 1. + ((params - 1.) / 4.)
        result = np.sin(np.pi * w[0])**2 + ((w[-1]-1) ** 2) * (1 + np.sin(2 * np.pi * w[-1]) ** 2)
        for i, x in enumerate(params):
            if i+1 == len(params):
                break
            result += ((w[i]-1)**2) * (1 + 10 * np.sin(np.pi * w[i] + 1) ** 2)

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

from olympus import Logger
from . import get_models_list
from . import import_model
from . import AbstractModel


def Model(kind='Neuralnet'):
	"""Convenience function to access planners via a slightly higher level interface It returns a certain planner
	with defaults arguments by keyword.

	Args:
		kind (str or AbstractModel): keyword identifying one of the models available in Olympus. Alternatively,
			you can pass a custom model that is a subclass of AbstractModel.

	Returns:
		model: an instance of the chosen model.
	"""
	_validate_model_kind(kind)
	# if a string is passed, then load the corresponding wrapper
	if type(kind) == str:
		model = import_model(kind)
		model = model()
	# if a custom class is passed, then that is the 'wrapper'
	elif issubclass(kind, AbstractModel):
		model = kind()

	return model
	

def _validate_model_kind(kind):
	# if we received a string
	if type(kind) == str:
		avail_models = get_models_list()
		if kind not in avail_models:
			message = ('Model "{0}" not available in Olympus. Please choose '
					   'from one of the available models: {1}'.format(kind, ', '.join(avail_models)))
			Logger.log(message, 'FATAL')

	# if we received a custom model class
	elif issubclass(kind, AbstractModel):
		# make sure it has the necessary methods
		for method in ['_train', '_predict']:
			implementation = getattr(kind, method, None)
			if not callable(implementation):
				message = f'The object {kind} does not implement the necessary method "{method}"'
				Logger.log(message, 'FATAL')

	# if we do not know what was passed raise an error
	else:
		message = 'Could not initialize Model: the argument "kind" is neither a string or AbstractModel subclass'
		Logger.log(message, 'FATAL')



#!/usr/bin/env python

#===============================================================================

import abc
import numpy as np

#===============================================================================

from olympus import Object

#===============================================================================

class AbstractPlotter(Object):


    @abc.abstractmethod
    def _plot(self, emulators, planners, measurements, *args, **kwargs):
        pass


    def plot_from_db(self, database, emulator=None, plot_file_name='test.png', *args, **kwargs):
        campaigns    = database.get_campaigns()
        emulators    = []
        planners     = []
        measurements = {}
        for campaign in campaigns:
            if campaign.emulator_type == 'n/a':  # this means campaign never got an Emulator or Surface
                continue
            emulator_type = campaign.get_emulator_type()
            if emulator_type == 'numeric':
                camp_emulator = f'{campaign.model_kind}_{campaign.dataset_kind}'
            elif emulator_type == 'analytic':
                camp_emulator = f'analytic_{campaign.surface_kind}'
            if emulator is None:
                emulator = camp_emulator
            if camp_emulator != emulator: continue
            planner  = campaign.get_planner_kind()
            if emulator not in emulators:
                emulators.append(emulator)
                measurements[emulator] = {}
            if planner not in planners:
                planners.append(planner)
                measurements[emulator][planner] = {'idxs': [], 'vals': []}
            for val_index, value in enumerate(campaign.best_values):
#                print('MEASUREMENTS', measurements[emulator].keys(), val_index, campaign.id)
#                try:
                measurements[emulator][planner]['idxs'].append(val_index)
                measurements[emulator][planner]['vals'].append(value)
#                except KeyError: continue

        for emulator in emulators:
            for planner in planners:
                try:
                    measurements[emulator][planner]['idxs'] = np.array(measurements[emulator][planner]['idxs'])
                    measurements[emulator][planner]['vals'] = np.array(measurements[emulator][planner]['vals'])
                except KeyError: continue

        self._plot(emulators, planners, measurements, file_name=plot_file_name, *args, **kwargs)



#!/usr/bin/env python

import numpy as np
from olympus.objects                   import ParameterVector
from olympus.planners.abstract_planner import AbstractPlanner
from olympus import Logger


class Grid(AbstractPlanner):

    def __init__(self, goal='minimize', levels=2, budget=None, exceed_budget=True, shuffle=False, random_seed=None):
        """Grid search.

        Note that the number of samples grow exponentially with the number of dimensions. E.g. for a 2-dimensional
        parameter space, with 2 levels, the grid will contain 4 samples; for a 3-dimensional space, it will contain
        8 samples; for a 6-dimensional space, 64 samples.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            levels (int, list): How many locations in parameter space to sample per parameter/dimension. If an integer
                is provided, the same number of levels will be used for each dimension. Provide a list to use different
                levels for different dimensions. Default is 2.
            budget (int): Maximum number of samples you want to evaluate. From the specified ``budget`` an adequate
                value for the levels will be derived. Note that, if provided, the argument ``levels`` will be discarded.
                Default is None.
            exceed_budget (bool): Whether to allow building the grid with more samples then ``budget``.
                This means some points might not be evaluated, but ensures there will be enough grid points to run as
                many evaluations as defined in ``budget``. If False, the number of grid points will be less or equal to
                ``budget``; this guarantees the budget is enough to guarantee the exploration of the whole grid.
                Default is True.
            shuffle (bool): Whether to randomize the order of the samples in the grid. Default is False.
            random_seed (int): Random seed. Set a random seed for a reproducible randomization of the grid if ``shuffle``
                was set to True.
        """
        AbstractPlanner.__init__(**locals())
        self.grid_created = False

    def reset(self):
        """Clears the remaining samples in the grid and prepares the planner for re-initialisation."""
        self.samples_loc = None
        self.samples = None
        self.grid_created = False

    def _set_param_space(self, param_space):
        self.param_space = param_space
        self.dims = len(self.param_space)
        self.bounds = self.param_space.param_bounds

        # if budget provided, define levels
        if self.budget is not None:
            self._get_approximate_levels()

        # allow providing a list of levels to tune the budget
        if isinstance(self.levels, int):
            self._levels = [self.levels] * self.dims
        elif isinstance(self.levels, list) or isinstance(self.levels, np.ndarray):
            if len(self.levels) != self.dims:
                message = (f'The number of levels provided ({len(self.levels)}) does not match dimensionality of the '
                           f'parameter space ({self.dimlists}).')
                Logger.log(message, 'ERROR')
            self._levels = list(self.levels)
        else:
            raise ValueError('Argument `level` can only be a integer or a list.')

    def _tell(self, observations):
        # grid search does not care about previous observations
        pass

    def _create_grid(self):
        self.samples_loc = []  # sample locations by dimension

        # define location of grid samples for each dimension
        for param, level in zip(self.param_space, self._levels):
            loc = np.linspace(start=param.low, stop=param.high, num=level)
            self.samples_loc.append(loc)

        meshgrid = np.stack(np.meshgrid(*self.samples_loc), len(self.samples_loc))  # make grid
        num_samples = np.prod(np.shape(meshgrid)[:-1])  # number of samples in grid

        # all grid samples in a 2D array
        self.samples = np.reshape(meshgrid, newshape=(num_samples, len(self.samples_loc)))

        # shuffle is we are sampling these points at random
        if self.shuffle is True:
            np.random.seed(self.random_seed)
            np.random.shuffle(self.samples)

        self.samples = list(self.samples)
        self.grid_created = True

    def _get_approximate_levels(self):
        # initialise with lowest number of samples that does not exceed budget
        base_level = int(np.floor(self.budget ** (1. / self.dims)))
        self.levels = np.array([base_level] * self.dims)
        effective_budget = np.prod(self.levels)  # actual number of samples created
        excess = int(self.budget - effective_budget)  # excess budget left

        # now add one level at a time to each dimension
        counter = 0
        while excess > 0:  # add levels until we exhaust all excess budget samples
            # if we want to have samples <= budget (i.e. do not exceed budget)
            if self.exceed_budget is False:
                # add 1 to last level as this should never get to receive a +1 anyway
                effective_budget_lookahead = np.prod(self.levels[:-1]) * (self.levels[-1] + 1)
                if effective_budget_lookahead > self.budget:
                    break
            # increment one level at a time
            i = counter % self.dims
            self.levels[i] += 1
            counter += 1
            # update effective budget and excess
            effective_budget = np.prod(self.levels)
            excess = int(self.budget - effective_budget)

        if excess >= 0:
            excess_type = 'budgeted evaluations'
        else:
            excess_type = 'grid points'
        message = (f'Budget provided, discarding argument `levels`. Given a budget of {self.budget} and a '
                   f'parameter space of dimensionality {self.dims}, we will explore {self.levels} levels. '
                   f'Given an excess of {np.abs(excess)} {excess_type}, the effective number '
                   f'of evaluations required is {effective_budget}.')
        Logger.log(message, 'INFO')

    def _ask(self):
        if self.grid_created is False:
            self._create_grid()
        param = self.samples.pop(0)

        if len(self.samples) == 0:
            message = 'Last parameter being provided - there will not be any more available samples in the grid.'
            Logger.log(message, 'INFO')

        return ParameterVector(array=param, param_space=self.param_space)






#!/usr/bin/env python

from . import AbstractPlanner


class CustomPlanner(AbstractPlanner):
    def __init__(self, goal='minimize'):
        """Parent class to be used to create custom Planner classes.

        Args:
            goal: The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
        """
        AbstractPlanner.__init__(**locals())

    def _set_param_space(self, param_space):
        self._param_space = []
        for param in param_space:
            if param.type == 'continuous':
                param_dict = {'name': param.name, 'type': param.type, 'domain': (param.low, param.high)}
                self._param_space.append(param_dict)

    def _tell(self, observations):
        self._params = observations.get_params(as_array=True)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)


#!/usr/bin/env python

# =====================
# Define Olympus colors
# =====================

import numpy as np
import matplotlib.pyplot as plt

from matplotlib.colors import LinearSegmentedColormap

_olympus_reference_colors = [
    "#08294C",
    "#75BBE1",
    "#D4E9F4",
    "#F2F2F2",
    "#F7A4D4",
    "#F75BB6",
    "#EB0789",
]
_olympus_cmap = LinearSegmentedColormap.from_list("olympus", _olympus_reference_colors)
_olympus_cmap_r = LinearSegmentedColormap.from_list(
    "olympus_r", _olympus_reference_colors[::-1]
)

plt.register_cmap(cmap=_olympus_cmap)
plt.register_cmap(cmap=_olympus_cmap_r)


def get_olympus_colors(n):
    olympus_cmap = plt.get_cmap("olympus")
    return [olympus_cmap(x) for x in np.linspace(0, 1, n)]



#!/usr/bin/env python 

import numpy as np 

#===============================================================================

def _get_init_guess_random(param_space, random_seed=None):
	np.random.seed(random_seed)
	init_guess = []
	for param in param_space:
		if param.type == 'continuous':
			init_guess.append(np.random.uniform(low=param.low, high=param.high))
	return np.array(init_guess)

#===============================================================================

def get_init_guess(param_space, method='random', **kwargs):
	if method == 'random':
		return _get_init_guess_random(param_space, **kwargs)
	else:
		raise NotImplementedError

#===============================================================================

def get_bounds(param_space):
	bounds = []
	for param in param_space:
		if param.type == 'continuous':
			bounds.append((param.low, param.high))
	return bounds

#===============================================================================



#!/usr/bin/env python

import numpy as np
from olympus.noises import AbstractNoise


class GaussianNoise(AbstractNoise):

    def __init__(self, scale=1):
        """Gaussian noise.

        Args:
            scale (float): Standard deviation of the Gaussian. Default is 1.
        """
        AbstractNoise.__init__(**locals())

    def _add_noise(self, value):
        noisy_value = np.random.normal(loc=value, scale=self.scale)
        return noisy_value



#!/usr/bin/env python

import os
import shutil
import argparse
import subprocess

# ===============================================================================

from olympus import __home__, Logger

# ===============================================================================


class ParserUpload:
    def __init__(self, subparsers):

        self.parser = subparsers.add_parser("upload", help=">> help for upload")
        self.parser.add_argument("-n", "--name", required=True)
        self.parser.add_argument("-p", "--path", default="./")
        self.parser.add_argument(
            "--no-fork", dest="fork", action="store_false",
        )
        self.parser.set_defaults(fork=True)

    @staticmethod
    def __call__(args):
        name = args.name
        path = args.path
        fork = args.fork

        # run some checks to make sure that the dataset is ok
        complete = True
        expected_files = ["dataset.zip", "data.csv", "description.txt"]
        for expected_file in expected_files:
            if not os.path.isfile(f"{path}/{expected_file}"):
                complete = False
                Logger.log(
                    f"could not find expected file {path}/{expected_file}", "ERROR"
                )
        if complete is False:
            Logger.log(
                "Please provide the dataset in the expected format and try again.",
                "INFO",
            )
            return

        # check that we have git
        git_path = shutil.which("git")
        if len(git_path) == 0:
            Logger.log(
                "Could not find a local version of git. Please install git or submit the pull request manually",
                "ERROR",
            )
            return

        # check that we have git-pull-request
        try:
            import git_pull_request
        except ModuleNotFoundError:
            Logger.log("Could not find a local version of git-pull-request", "ERROR")
            return

        # Inspired from the link below, we run a couple of git commands
        # https://medium.com/mergify/managing-your-github-pull-request-from-the-command-line-89cb6af0a7fa
        Logger.log("uploading dataset", "INFO")

        with open(f"{__home__}/cli/template_push_to_github.sh", "r") as content:
            template = content.read()

        replace_dict = {
            "{@DATASET_NAME}": name,
            "{@PATH}": f"../{path}",
            "{@NO_FORK}": "" if fork is True else "--no-fork",
        }

        for key, value in replace_dict.items():
            template = template.replace(str(key), str(value))

        push_file_name = "./push_to_github.sh"
        with open(push_file_name, "w") as content:
            content.write(template)

        # get default git editor
        editor_file = ".editor"
        subprocess.call(f"git config --get core.editor > {editor_file}", shell=True)
        with open(editor_file, "r") as content:
            editor = content.read().strip()
        subprocess.call(
            f"bash push_to_github.sh < {__home__}/cli/template_sign_message_{editor}.sh",
            shell=True,
        )

        # remove push script
        os.remove(push_file_name)


# ===============================================================================



#!/usr/bin/env python

#===============================================================================

import numpy as np

from olympus         import Logger
from olympus.objects import ObjectParameter
from olympus.objects import ParameterContinuous

#===============================================================================

class ParameterSpace:
    ''' generic parameter space

    This class is intended to store all information about a parameter
    space subject to optimization, provide the interface to define a
    parameter space and perform basic operations on the parameter
    space (constraints, boundaries, etc.)
    '''

    def __init__(self):
        self.parameters = []

    def __getitem__(self, index):
        return self.parameters[index]

    def __iter__(self):
        for param in self.parameters:
            yield param

    def __len__(self):
        return len(self.parameters)

    def __str__(self):
        return '\n'.join([str(param) for param in self])

    @property
    def param_names(self):
        return [param.name for param in self]

    @property
    def param_bounds(self):
        return [(param.low, param.high) for param in self]

    @property
    def param_lowers(self):
        return np.array([param.low for param in self])

    @property
    def param_uppers(self):
        return np.array([param.high for param in self])

    def _add_param(self, param):
        # check if we already have that param
        if param.name in self.param_names:
            message = '''Parameter "{}" is already defined'''.format(param.name)
            Logger.log(message, 'ERROR')
        else:
            self.parameters.append(param)

    def add(self, param):
        """

        Args:
            param:

        Returns:

        """
        if isinstance(param, ObjectParameter):
            self._add_param(param)
        elif isinstance(param, list):
            for _param in param:
                self._add_param(_param)
        else:
            Logger.log('Please provide a valid parameter', 'ERROR')

    def get_param(self, name):
        """

        Args:
            name:

        Returns:

        """
        for param in self.parameters:
            if param['name'] == name:
                return param
        message = 'Could not find Parameter with name {0} in {1}'.format(name, str(self))
        Logger.log(message, 'WARNING')
        return None

    def item(self):
        """

        Returns:

        """
        return [param.to_dict() for param in self]

    def validate(self, param_vector):
        """

        Args:
            param_vector:

        Returns:

        """
        bounds = self.param_bounds
        valid  = True
        for i, bound in enumerate(bounds):
            entry = param_vector[i]
            valid = valid and bound[0] <= entry
            valid = valid and entry <= bound[1]

        if valid is False:
            message = f'Not all parameters of {param_vector} are within bounds!'
            Logger.log(message, 'WARNING')

        return valid



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface


class Rosenbrock(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """Rosenbrock function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        # minimum at (1,1) for 2d
        # vector = np.array([1,1])
        # print((vector+2)/4)
        # minimum at the centre
        params = [0.75] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        return None

    def _run(self, params):
        params = np.array(params)
        params = 4 * params - 2  # rescale onto [-2, 2]
        result = 0
        for index, element in enumerate(params[:-1]):
            result += 100 * (params[index + 1] - element ** 2) ** 2 + (1 - element) ** 2

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

from olympus.planners import AbstractPlanner
from olympus.objects import ParameterVector


class Gpyopt(AbstractPlanner):

    def __init__(self, goal='minimize', batch_size=1, exact_eval=True, model_type='GP_MCMC', acquisition_type='EI_MCMC'):
        """
        Gaussian Process optimization as implemented in GPyOpt.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            batch_size (int): size of the batch in which the objective is evaluated.
            exact_eval (bool): whether the outputs are exact.
            model_type (str): type of model to use as surrogate. 'GP': standard Gaussian process. 'GP_MCMC': Gaussian
                process with prior in the hyper-parameters. 'sparseGP': sparse Gaussian process. 'warperdGP': warped
                Gaussian process. 'InputWarpedGP': input warped Gaussian process. 'RF': random forest (scikit-learn).
            acquisition_type (str): type of acquisition function to use. 'EI': expected improvement. 'EI_MCMC': integrated
                expected improvement (requires GP_MCMC model). 'MPI': maximum probability of improvement. 'MPI_MCMC':
                maximum probability of improvement (requires GP_MCMC model). 'LCB': GP-Lower confidence bound. 'LCB_MCMC':
                integrated GP-Lower confidence bound (requires GP_MCMC model).
        """
        AbstractPlanner.__init__(**locals())

    def _set_param_space(self, param_space):
        self._param_space = []
        for param in param_space:
            if param.type == 'continuous':
                param_dict = {'name': param.name, 'type': param.type, 'domain': (param.low, param.high)}
            self._param_space.append(param_dict)

    def _tell(self, observations):
        self._params = observations.get_params(as_array = 'True')
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)

    def _get_bo_instance(self):
        from GPyOpt.methods import BayesianOptimization
        bo = BayesianOptimization(
                f                = lambda x: 0, ### TODO: NEED TO CHECK IF THIS CAUSES ISSUES
                domain           = self._param_space,
                batch_size       = self.batch_size,
                exact_eval       = self.exact_eval,
                model_type       = self.model_type,
                acquisition_type = self.acquisition_type,
                X = self._params, Y = self._values,
            )
        return bo

    def _ask(self):
        if self._params is None or len(self._params) < 2:
            self._params = None
            self._values = None
            bo = self._get_bo_instance()
            generated = bo.get_evaluations()[0]
            array = generated[self.num_generated % len(generated)]
        else:
            bo    = self._get_bo_instance()
            array = bo.suggest_next_locations()[0]
        return ParameterVector().from_array(array, self.param_space)



#!/usr/bin/env python

import os
import traceback

__home__ = os.path.dirname(os.path.abspath(__file__))

# ===============================================================================

from olympus import Logger

# ===============================================================================

# check for pandas
try:
    import pandas
except ModuleNotFoundError:
    error = traceback.format_exc()
    for line in error.split("\n"):
        if "ModuleNotFoundError" in line:
            module = line.strip().strip("'").split("'")[-1]
    message = """Datasets requires {module}, which could not be found.
    Install {module} for accessing datasets directly or providing custom
    datasets""".format(
        module=module
    )
    Logger.log(message, "FATAL")
finally:
    from .dataset import Dataset, load_dataset

# ===============================================================================

import glob

datasets_list = []
for dir_name in glob.glob(f"{__home__}/dataset_*"):
    dir_name = dir_name.split("/")[-1][8:]
    datasets_list.append(dir_name)


def list_datasets():
    return sorted(datasets_list).copy()



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface
from itertools import product


class HyperEllipsoid(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """Hyper-ellipsoid function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        # minimum at the centre
        params = [0.5] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        # maxima at the corners
        maxima = []
        params = product([0, 1], repeat=self.param_dim)
        for param in params:
            param = list(param)
            value = self._run(param)
            maxima.append({'params': param, 'value': value})
        return maxima

    def _run(self, params):
        params = np.array(params)
        params = 10 * params - 5  # rescale onto [-5, 5]
        weights = np.arange(1, len(params) + 1)
        result = 0
        for index, element in enumerate(params):
            result += weights[index] * element ** 2

        if self.noise is None:
            return result
        else:
            return self.noise(result)


#!/usr/bin/env python

#======================================================================

import os
__home__ = os.path.dirname(os.path.abspath(__file__))

#======================================================================

#from .cv_manager         import CV_Manager
from .emulator           import Emulator

# get list of all available emulators
def list_trained_emulators():
    import glob
    emulators = []
    for dir_name in glob.glob('{}/emulator_*'.format(__home__)):
        if os.path.isfile(dir_name): continue
        dataset_name = dir_name.split('_')[-2]
        model_name   = dir_name.split('_')[-1]
        emulators.append(f'{dataset_name}_{model_name}')
    return emulators

#======================================================================



#!/usr/bin/env python

from olympus.models import WrapperTensorflowModel

#===============================================================================

import numpy as np
import time
import tensorflow as tf

#===============================================================================

class NeuralNet(WrapperTensorflowModel):

    ATT_KIND = {'type': 'string', 'default': 'NeuralNet'}

    def __init__(self, scope='model', hidden_depth=3, hidden_nodes=48, hidden_act='leaky_relu', out_act='linear',
                 l2_activity=1e-3, gaussian_dropout=0.0, dropout=0.1,
                 learning_rate=1e-3, pred_int=100,  reg=0.001, es_patience=100, max_epochs=100000, batch_size=20):
        '''Neural network model.

        Args:
            scope (str): TensorFlow scope.
            hidden_depth (int): Number of hidden layers.
            hidden_nodes (int): Number of hidden nodes per layer.
            hidden_act (str): Hidden activation function. Available options are 'linear', 'leaky_relu', 'relu',
                'softmax', 'softplus', 'softsign', 'sigmoid'.
            out_act (str): Output activation function. Available options are 'linear', 'leaky_relu', 'relu',
                'softmax', 'softplus', 'softsign', 'sigmoid'.
            l2_activity (float): L2 regularization.
            learning_rate (float): Learning rate.
            pred_int (int): Frequency with which we make predictions on the validation/training set (in number of epochs).
            reg (float): ???
            es_patience (int): Early stopping patience.
            max_epochs (int): Maximum number of epochs allowed.
            batch_size (int): Size batches used for training.
        '''

        WrapperTensorflowModel.__init__(**locals())

    def _build_inference(self):

        self.graph = tf.Graph()
        self.is_graph_constructed = True

        activation_hidden = self.act_funcs(self.hidden_act)
        activation_out    = self.act_funcs(self.out_act)

        STDDEV = 1e-1

        with self.graph.as_default():
            self.tf_x = tf.compat.v1.placeholder(tf.float32, [self.batch_size, self.features_dim])
            self.tf_y = tf.compat.v1.placeholder(tf.float32, [self.batch_size, self.targets_dim])

            with tf.name_scope(self.scope):

                self.layers = []
                for _ in range(self.hidden_depth):
                    self.layers.append(tf.keras.layers.BatchNormalization())
                    self.layers.append(tf.keras.layers.Dense(self.hidden_nodes,
                        activation=activation_hidden,
                        kernel_regularizer=tf.keras.regularizers.l2(l = self.l2_activity)))
                    self.layers.append(tf.keras.layers.GaussianDropout(self.gaussian_dropout))
                    self.layers.append(tf.keras.layers.Dropout(self.dropout))

                self.layers.append(tf.keras.layers.BatchNormalization())
                self.layers.append(tf.keras.layers.Dense(self.targets_dim,
                                               activation = activation_out,
                                               kernel_regularizer = tf.keras.regularizers.l2(l = self.l2_activity)))

                self.neural_net = tf.keras.Sequential(self.layers)
                self.y_pred     = self.neural_net(self.tf_x)

                reg_vars = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.REGULARIZATION_LOSSES)
                self.reg_loss   = self.reg * tf.reduce_mean(reg_vars)
                self.pred_loss  = tf.sqrt(tf.reduce_mean(tf.square(self.y_pred - self.tf_y)))
                self.pred_loss += tf.reduce_mean(tf.abs(self.y_pred - self.tf_y))
                self.loss       = self.pred_loss + self.reg_loss

                self.optimizer = tf.compat.v1.train.AdamOptimizer(self.learning_rate)
                self.train_op  = self.optimizer.minimize(self.loss)

                self.init_op   = tf.group(tf.compat.v1.global_variables_initializer(), tf.compat.v1.local_variables_initializer())
                self.sess = tf.compat.v1.Session(graph = self.graph)
                with self.sess.as_default():
                    self.sess.run(self.init_op)
                    self.saver = tf.compat.v1.train.Saver()

#===============================================================================



#!/usr/bin/env python

import numpy as np

from olympus.planners              import AbstractPlanner
from olympus.objects               import ParameterVector
from olympus.objects.object_config import Config


class Phoenics(AbstractPlanner):

    def __init__(self, goal='minimize', batches=1, boosted=True, parallel=True, sampling_strategies=2):
        """
        A Bayesian optimization algorithm based on Bayesian Kernel Density estimation.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            batches (int): number of parameter batches to return at each 'ask'
                iteration.
            boosted (bool): whether to use a lower fidelity approximation in
                regions of low density during kernel density estimation. Setting
                this to True reduces the run time of the planner.
            parallel (bool): whether to run the code in parallel.
            sampling_strategies (int): number of sampling strategies to use.
                Each sampling strategy uses a different balance between
                exploration and exploitation.
        """
        AbstractPlanner.__init__(**locals())

        self._observations = []
        self._counter = 0

    def _set_param_space(self, param_space):
        self._param_space = []
        for param in param_space:
            if param.type == 'continuous':
                param_dict = {'name': param.name, 'type': param.type, 'size': 1, 'low': param.low, 'high': param.high}
            else:
                raise NotImplementedError
            self._param_space.append(param_dict)

    def _tell(self, observations):
        self._params       = observations.get_params()
        self._values       = observations.get_values(as_array=True, opposite=self.flip_measurements)
        self._observations = []
        for _, param in enumerate(self._params):
            obs = ParameterVector().from_array(param, self.param_space).to_dict()
            for key, value in obs.items():
                obs[key] = np.array([value])
            obs['obj'] = self._values[_]
            # WARNING: THIS IS JUST A HACK
            # NOTE: we should make sure we are not adding unnecessary dimensions to this array
            obs['obj'] = np.squeeze(obs['obj'])
            self._observations.append(obs)

    def _get_phoenics_instance(self):
        from phoenics import Phoenics as ActualPhoenics
        params = self.param_space.item()
        for param in params:
            param['size'] = 1
        config_dict = {
            'general': self.config.to_dict(),
            'parameters': params,
            'objectives': [{'name': 'obj', 'goal': 'minimize'}]}
        self.phoenics = ActualPhoenics(config_dict=config_dict)

    def _ask(self):
        self._get_phoenics_instance()
        params = self.phoenics.recommend(observations=self._observations)
        # check which param to return
        param = params[self._counter % len(params)]
        for key, value in param.items():
            param[key] = np.squeeze(value)
        self._counter += 1
        param = ParameterVector().from_dict(param)
        return param



#!/usr/bin/env python  

#==============================================================================

import traceback

from olympus import Logger 

#==============================================================================

try:
	import sqlalchemy
except ModuleNotFoundError:
	error = traceback.format_exc()
	for line in error.split('\n'):
		if 'ModuleNotFoundError' in line:
			module = line.strip().strip("'").split("'")[-1]
	message = '''Sqlite databases require {module}, which could not be found. 
	Please install {module} or use a different database backend'''.format(module = module)
	Logger.log(message, 'WARNING', only_once=True)
	raise ModuleNotFoundError

#==============================================================================

from olympus.databases.database_sqlite.sqlite_operations import AddEntry, UpdateEntries, FetchEntries
from olympus.databases.database_sqlite.sqlite_interface  import SqliteInterface
from olympus.databases.database_sqlite.wrapper_sqlite    import Wrapper_sqlite



#!/usr/bin/env python

import os
import numpy as np
import pickle
import shutil
from glob import glob
from copy import deepcopy
from tempfile import TemporaryDirectory
from sklearn.metrics import r2_score

from olympus import __emulator_path__, __scratch__, __version__, Logger
from olympus.datasets.dataset import Dataset
from olympus.models.model import Model
from olympus.models.abstract_model import AbstractModel
from olympus.objects import Object, ParameterVector
from olympus.utils.data_transformer import DataTransformer
from olympus.models.model import _validate_model_kind
from olympus.datasets.dataset import _validate_dataset_args


# =========================
# Main Class of This Module
# =========================
class Emulator(Object):
    """ generic experiment emulator

    This class is intended to provide the interface to the user.

    Random notes:
    - emulators are uniquely determined via dataset + model + emulator_id
    """

    def __init__(
        self,
        dataset=None,
        model=None,
        feature_transform="identity",
        target_transform="identity",
    ):
        """Experiment emulator.

        Args:
            dataset (str, Dataset): dataset used to train a model. Either a string, in which case a standard dataset
                is loaded, or a Dataset object. To see the list of available datasets ...
            model (str, Model): the model used to create the emulator. Either a string, in which case a default model
                is loaded, or a Model object. To see the list of available models ...
            feature_transform (str, list): the data transform to be applied to the features. See DataTransformer for the
                available transformations.
            target_transform (str, list): the data transform to be applied to the targets. See DataTransformer for the
                available transformations.
        """

        # ------------------------------------------------------------
        # if dataset and model are strings ==> load emulator from file
        # ------------------------------------------------------------
        if type(dataset) == str and type(model) == str:
            # check dataset string
            _validate_dataset_args(
                kind=dataset, data=None, columns=None, target_names=None
            )
            # check model string
            _validate_model_kind(model)
            Logger.log(
                f"Loading emulator using a {model} model for the dataset {dataset}...",
                "INFO",
            )
            self._load(f"{__emulator_path__}/emulator_{dataset}_{model}")

        # -----------------------------------------
        # otherwise, assume it is a custom emulator
        # -----------------------------------------
        else:
            Object.__init__(**locals())

            if dataset is not None:
                self._set_dataset(dataset)
            if model is not None:
                self._set_model(model)

            # other attributes we will use
            self._version = __version__
            self._ghost_model = deepcopy(self.model)
            self.is_trained = False
            self.cross_val_performed = False
            self.cv_scores = None
            self.model_scores = None
            self.emulator_to_save = None
            self.feature_transformer = DataTransformer(
                transformations=self.feature_transform
            )
            self.target_transformer = DataTransformer(
                transformations=self.target_transform
            )

        # create tmp dir to store model files
        # also if we are loading a model (the user could call 'train' again)
        self._scratch_dir = TemporaryDirectory(dir=f"{__scratch__}", prefix="emulator_")

    def __str__(self):
        if self.dataset is not None and self.model is not None:
            return f"<Emulator ({self.dataset}, model={self.model})>"
        elif self.dataset is not None:
            return f"<Emulator ({self.dataset})>"
        elif self.model is not None:
            return f"<Emulator (model={self.model})>"
        else:
            return f"<Emulator (unspecified)>"

    @property
    def goal(self):
        return self.dataset.goal

    @property
    def param_space(self):
        return self.dataset.param_space

    @property
    def value_space(self):
        return self.dataset.value_space

    # ===========
    # Set Methods
    # ===========
    def _set_dataset(self, dataset):
        """ registers a dataset for emulator

        Args:
            dataset (str): name of available dataset, or Dataset object.
        """
        if type(dataset) == str:
            self.dataset = Dataset(kind=dataset)
        elif isinstance(dataset, Dataset):
            self.dataset = dataset
        else:
            raise NotImplementedError

        # check that the param_space is defined
        if self.dataset.param_space is None:
            message = (
                "The param_space information is not present in the Dataset object provided. Please use "
                "Dataset.set_param_space to define the type of variables present in the dataset before "
                "instantiating the Emulator."
            )
            Logger.log(message, "ERROR")

    def _set_model(self, model):
        """ registers a model for emulator

        Args:
            model (str): name of available model, or a model object
        """
        if type(model) == str:
            self.model = Model(kind=model)
        elif isinstance(model, AbstractModel):
            self.model = model
        else:
            raise NotImplementedError
        # self.param_space is taken from self.dataset.param_space, so...
        if self.dataset is None:
            self.model.set_param_space(None)
        else:
            self.model.set_param_space(self.param_space)

    # =========================
    # Train and Predict Methods
    # =========================
    def cross_validate(self, rerun=False, plot=False):
        # TODO: allow setting verbosity: verbose=True/False will be enough
        """Performs cross validation on the emulator dataset, using the emulator model. The number of folds used is
        defined in the Dataset object.

        Args:
            rerun (bool): whether to run cross validation again, in case it had already been performed.

        Returns:
            scores (dict): dictionary with the list of train and validation R2 scores.

        """
        """Perform cross validation.

        Returns (dict): dictionary containing the training and test R2 scores for all folds.

        """

        if self.cross_val_performed is True and rerun is False:
            message = (
                "Cross validation has already been performed for this Emulator. You can see its results in "
                "`self.cv_scores`. If you would like to rerun cross validation and overwrite the previous "
                "results, set `rerun` to True"
            )
            Logger.log(message, "FATAL")

        training_r2_scores = np.empty(self.dataset.num_folds)
        valid_r2_scores = np.empty(self.dataset.num_folds)
        training_rmsd_scores = np.empty(self.dataset.num_folds)
        valid_rmsd_scores = np.empty(self.dataset.num_folds)

        # get scaled train/valid sets
        # NOTE: we do not want to use the self.transformers, because for 'run' we want to use the transformers
        # trained in 'train'. If we reset the Transformers here, then if a user calls 'cross_validate' after 'train'
        # we end up using the wrong transformers in 'run'
        feature_transformer = DataTransformer(transformations=self.feature_transform)
        target_transformer = DataTransformer(transformations=self.target_transform)

        # ---------------------------------------
        # Iterate over the cross validation folds
        # ---------------------------------------
        for fold in range(self.dataset.num_folds):
            # get the train/valid sets
            # NOTE: we keep the features as Dataset objects, as these are needed for possible periodic transformations
            # TODO: expend the above also to targets? Right now param_space does not describe what type of variable
            #  the targets are
            train_features = Dataset(data=self.dataset.cross_val_sets_features[fold][0])
            train_features.set_param_space(self.dataset.param_space)
            valid_features = self.dataset.cross_val_sets_features[fold][1].to_numpy()
            train_targets = self.dataset.cross_val_sets_targets[fold][0].to_numpy()
            valid_targets = self.dataset.cross_val_sets_targets[fold][1].to_numpy()

            feature_transformer.train(train_features)
            target_transformer.train(train_targets)

            train_features_scaled = feature_transformer.transform(train_features)
            valid_features_scaled = feature_transformer.transform(valid_features)
            train_targets_scaled = target_transformer.transform(train_targets)
            valid_targets_scaled = target_transformer.transform(valid_targets)

            # define scope and make a copy of the model for the cross validation
            model_fold = deepcopy(
                self._ghost_model
            )  # the model we will use for training the fold
            model_fold.scope = f"Fold_{fold}"
            model_path = f"{self._scratch_dir.name}/{model_fold.scope}"
            # TODO/QUESTION: in case we are overwriting the output of a previous call, should we first remove the folder
            #  to make sure to have a cleared path?
            if not os.path.exists(model_path):
                os.makedirs(model_path)

            Logger.log(f">>> Training model on fold #{fold}...", "INFO")
            (
                mdl_train_r2,
                mdl_valid_r2,
                mdl_train_rmsd,
                mdl_test_rmsd,
            ) = model_fold.train(
                train_features=train_features_scaled,
                train_targets=train_targets_scaled,
                valid_features=valid_features_scaled,
                valid_targets=valid_targets_scaled,
                model_path=model_path,
                plot=plot,
            )

            # store performance of fold
            training_r2_scores[fold] = mdl_train_r2
            valid_r2_scores[fold] = mdl_valid_r2
            training_rmsd_scores[fold] = mdl_train_rmsd
            valid_rmsd_scores[fold] = mdl_test_rmsd
            # write file to indicate training is complete and add R2 in there
            with open(f"{model_path}/training_completed.info", "w") as content:
                content.write(
                    f"Train R2={mdl_train_r2}\nValidation R2={mdl_valid_r2}\n"
                    f"Train RMSD={mdl_train_rmsd}\nValidation RMSD={mdl_test_rmsd}\n"
                )

        # print some info to screen
        Logger.log(f"Performance statistics based on transformed data "
                   f"[{self.feature_transform}, {self.target_transform}]:", "INFO")
        cv_r2_score_mean = np.mean(valid_r2_scores)
        cv_r2_score_stderr = np.std(valid_r2_scores) / np.sqrt(
            (len(valid_r2_scores) - 1)
        )
        cv_rmsd_score_mean = np.mean(valid_rmsd_scores)
        cv_rmsd_score_stderr = np.std(valid_rmsd_scores) / np.sqrt(
            (len(valid_rmsd_scores) - 1)
        )
        Logger.log(
            "Validation   R2: {0:.4f} +/- {1:.4f}".format(
                cv_r2_score_mean, cv_r2_score_stderr
            ),
            "INFO",
        )
        Logger.log(
            "Validation RSMD: {0:.4f} +/- {1:.4f}".format(
                cv_rmsd_score_mean, cv_rmsd_score_stderr
            ),
            "INFO",
        )

        self.cross_val_performed = True
        self.cv_scores = {
            "train_r2": training_r2_scores,
            "validate_r2": valid_r2_scores,
            "train_rmsd": training_rmsd_scores,
            "validate_rmsd": valid_rmsd_scores,
        }
        return self.cv_scores

    def train(self, plot=False, retrain=False):
        # TODO: allow setting verbosity: verbose=True/False will be enough
        """Trains the model on the emulator dataset, using the emulator model. The train/test split is defined in the
        Dataset object `emulator.dataset`. Note that the test set is used for testing the model performance, and for
        early stopping.

        Args:
            plot (bool):
            retrain (bool): whether to retrain the model, in case it had already been trained.

        Returns:
            scores (dict): dictionary with the train and test R2 scores.

        """
        # check if this emulator has already been trained
        if self.is_trained is True and retrain is False:
            message = (
                "The Emulator is already trained. If you would like to overwrite the already trained emulator, "
                "set `retrain` to True"
            )
            Logger.log(message, "FATAL")

        # get the train/test sets
        # NOTE: we keep the features as Dataset objects, as these are needed for possible periodic transformations
        # TODO: expend the above also to targets? Right now param_space does not describe what type of variable
        #  the targets are
        train_features = Dataset(data=self.dataset.train_set_features)
        train_features.set_param_space(self.dataset.param_space)
        test_features = self.dataset.test_set_features.to_numpy()
        train_targets = self.dataset.train_set_targets.to_numpy()
        test_targets = self.dataset.test_set_targets.to_numpy()

        # get scaled train/valid sets. These are also the DataTransformer objects we keep as they are needed in 'run'
        self.feature_transformer.train(train_features)
        self.target_transformer.train(train_targets)

        train_features_scaled = self.feature_transformer.transform(train_features)
        test_features_scaled = self.feature_transformer.transform(test_features)
        train_targets_scaled = self.target_transformer.transform(train_targets)
        test_targets_scaled = self.target_transformer.transform(test_targets)

        # define path where to store model
        model_path = f"{self._scratch_dir.name}/Model"
        # TODO/QUESTION: in case we are overwriting the output of a previous call, should we first remove the folder
        #  to make sure to have a cleared path?
        if not os.path.exists(model_path):
            os.makedirs(model_path)

        # Train
        Logger.log(
            ">>> Training model on {0:.0%} of the dataset, testing on {1:.0%}...".format(
                (1 - self.dataset.test_frac), self.dataset.test_frac
            ),
            "INFO",
        )
        mdl_train_r2, mdl_test_r2, mdl_train_rmsd, mdl_test_rmsd = self.model.train(
            train_features=train_features_scaled,
            train_targets=train_targets_scaled,
            valid_features=test_features_scaled,
            valid_targets=test_targets_scaled,
            model_path=model_path,
            plot=plot,
        )

        # write file to indicate training is complete and add R2 in there
        with open(f"{model_path}/training_completed.info", "w") as content:
            content.write(
                f"Train R2={mdl_train_r2}\nValidation R2={mdl_test_r2}\n"
                f"Train RMSD={mdl_train_rmsd}\nValidation RMSD={mdl_test_rmsd}\n"
            )

        Logger.log(f"Performance statistics based on transformed data "
                   f"[{self.feature_transform}, {self.target_transform}]:", "INFO")
        Logger.log("Train R2   Score: {0:.4f}".format(mdl_train_r2), "INFO")
        Logger.log("Test  R2   Score: {0:.4f}".format(mdl_test_r2), "INFO")
        Logger.log("Train RMSD Score: {0:.4f}".format(mdl_train_rmsd), "INFO")
        Logger.log("Test  RMSD Score: {0:.4f}\n".format(mdl_test_rmsd), "INFO")

        # set is_trained to True
        self.is_trained = True

        # show stats on untransformed samples
        # -----------------------------------
        y_train = self.dataset.train_set_targets.to_numpy()
        y_train_pred = self.run(features=self.dataset.train_set_features.to_numpy(), num_samples=10)
        y_test = self.dataset.test_set_targets.to_numpy()
        y_test_pred = self.run(features=self.dataset.test_set_features.to_numpy(), num_samples=10)
        train_r2 = r2_score(y_train, y_train_pred)
        train_rmse = np.sqrt(np.mean((y_train.flatten() - y_train_pred.flatten()) ** 2))
        test_r2 = r2_score(y_test, y_test_pred)
        test_rmse = np.sqrt(np.mean((y_test.flatten() - y_test_pred.flatten()) ** 2))

        Logger.log(f"Performance statistics based on original data:", "INFO")
        Logger.log("Train R2   Score: {0:.4f}".format(train_r2), "INFO")
        Logger.log("Test  R2   Score: {0:.4f}".format(test_r2), "INFO")
        Logger.log("Train RMSD Score: {0:.4f}".format(train_rmse), "INFO")
        Logger.log("Test  RMSD Score: {0:.4f}\n".format(test_rmse), "INFO")

        # save and return scores
        self.model_scores = {
            "train_r2": mdl_train_r2,
            "test_r2": mdl_test_r2,
            "train_rmsd": mdl_train_rmsd,
            "test_rmsd": mdl_test_rmsd,
        }
        return self.model_scores

    def run(self, features, num_samples=1, return_paramvector=False):
        """Run the emulator and return a value given the features provided.

        Args:
            features (ndarray): 2d array with the input features used for the predictions
            num_samples (int): number of samples to average. only useful for probabilistic models
            return_paramvector (bool): Whether to return a ``ParameterVector`` object instead of a list of lists.
                Default is False.

        Returns:
            y (float or array): model prediction(s).
        """

        # first check if we have a trained model
        if not self.is_trained:
            message = "This emulator has not been trained yet. Please train the emulator before you can use it for prediction."
            Logger.log(message, "ERROR")

        # check the inputs
        if type(features) == list:
            features = np.array(features)
        if len(features.shape) == 1:
            features = np.expand_dims(features, axis=0)

        # validate features
        if not features.shape[1] == len(self.param_space):
            message = (
                "Dimensions of provided features (%d) did not match expected dimension (%d)"
                % (features.shape[1], len(self.param_space))
            )
            Logger.log(message, "ERROR")
        for feature in features:
            if not self.param_space.validate(feature):
                message = "Not all parameters are within bounds"
                Logger.log(message, "WARNING")

        # scale the features using the DataTransformer that was fit when training
        features_scaled = self.feature_transformer.transform(features)

        # predict, drawing a certain amount of samples
        y_pred_scaled = self.model.predict(features_scaled, num_samples=num_samples)

        # return the prediction after inverting the transform
        y_preds = self.target_transformer.back_transform(
            y_pred_scaled
        )  # this is a 2d array

        # if we are not asking for a ParamVector, we can just return y_preds
        if return_paramvector is False:
            return y_preds

        # NOTE: while we do not allow batches or multiple objectives yet, this code is supposed to be able to support
        #  those
        y_pred_objects = []  # list of ParamVectors with all samples and objectives
        # iterate over all samples (if we returned a batch of predictions)
        for y_pred in y_preds:
            y_pred_object = ParameterVector()
            # iterate over all objectives/targets
            for target_name, y in zip(self.dataset.target_names, y_pred):
                y_pred_object.from_dict({target_name: y})
            # append object to list
            y_pred_objects.append(y_pred_object)

        return y_pred_objects

    def save(self, path="./olympus_emulator", include_cv=False):
        """Save the emulator in a specified location. This will save the emulator object as a pickle file, and the
        associated TensorFlow model, in the specified location. The saved emulator can then be loaded with the
        `olympus.emulators.load_emulator` function.

        Args:
            path (str): relative path where to save the emulator.
            include_cv (bool): whether to include the cross validation models. Default is False.

        """

        # TODO: check if path corresponds to one of our emulators and in that case raise error?

        # create path, or overwrite existing one
        if os.path.exists(path):
            message = f"Overwriting existing emulator in {path}!"
            Logger.log(message, "WARNING")
            shutil.rmtree(path)
        os.makedirs(path)

        self.emulator_to_save = Emulator()
        for key in self.__dict__:
#            skip = [
                #                'me',
                #                'indent',
                #                'props',
                #                'attrs',
                #                'max_prop_len',
                #                'dataset',
                #                'model',
                #                'feature_transform',
                #                'target_transform',
                #                '_version',
                #                '_ghost_model',
                #                'is_trained',
                #                'cross_val_performed',
                #                'cv_scores',
                #                'model_scores',
                #                'emulator_to_save',
                #                'feature_transformer',
                #                'target_transformer',
                #                '_scratch_dir']
#            ]
#            if key in skip:
#                continue

            if key == "model":
                setattr(self.emulator_to_save, key, self._ghost_model)
            elif key != "emulator_to_save":
                setattr(self.emulator_to_save, key, self.__dict__[key])
        self.emulator_to_save.reset()

        # pickle emulator object
        with open(f"{path}/emulator.pickle", "wb") as f:
            pickle.dump(self.emulator_to_save, f)

        # copy over model files
        if self.is_trained is True:
            shutil.copytree(f"{self._scratch_dir.name}/Model", f"{path}/Model")
        else:
            message = "The emulator you are saving has not been trained. Its model-related files will not be written to disk."
            Logger.log(message, "WARNING")

        # if include_cv is True ==> copy also CV models
        if include_cv is True:
            if self.cross_val_performed is True:
                for folder in glob(f"{self._scratch_dir.name}/Fold*"):
                    dirname = folder.split("/")[-1]
                    shutil.copytree(folder, f"{path}/{dirname}")
            else:
                message = (
                    "You are trying to save the cross validation models for an emulator that did not perform "
                    "cross validation. No cross validation models to be saved."
                )
                Logger.log(message, "WARNING")

        # TODO: also dump some info in human readable format? json like we had before?

    def _load(self, emulator_folder):
        emulator = load_emulator(emulator_folder)
        self.__dict__.update(emulator.__dict__)


# ===============
# Other Functions
# ===============
def load_emulator(emulator_folder):
    """Loads a previously saved emulator.

    Args:
        emulator_folder (str): path to the folder where the emulator was saved.

    Returns:
        emulator (Emulator): emulator object loaded from file.

    """

    # check path exists
    if os.path.exists(emulator_folder) is False:
        message = f'Folder "{emulator_folder}" does not exist'
        Logger.log(message, "FATAL")

    try:
        with open(f"{emulator_folder}/emulator.pickle", "rb") as f:
            emulator_stored = pickle.load(f)
    except TypeError:
        Logger.log("Failed to restore emulator", "FATAL")

    emulator_to_load = Emulator()
    for key in emulator_stored.__dict__:
        if key in emulator_to_load.__dict__:
            emulator_to_load.add(key, emulator_stored.get(key))
        else:
            message = (
                f'Key "{key}" from Emulator of olympus v.{emulator_stored._version} not found in Emulator '
                f'class of olympus v.{__version__}). Attribute "{key}" will not be set.'
            )
            Logger.log(message, "WARNING")

    if emulator_to_load.is_trained is True:
        emulator_to_load.model._set_dims(
            features_dim=emulator_to_load.dataset.features_dim,
            targets_dim=emulator_to_load.dataset.targets_dim,
        )
        restored = emulator_to_load.model.restore(f"{emulator_folder}/Model")
        if restored is False:
            message = "failed to restore model"
            Logger.log(message, "ERROR")

    # NOTE: we are not leading any cross validation models for the moment. If CV was performed though, we still have
    # the data about the CV performance
    return emulator_to_load



#!/usr/bin/env python

import os
import sys
import glob

__home__ = os.path.dirname(os.path.abspath(__file__))


class NoiseLoader:

    def __init__(self, *args, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        self._find_noise()

    def __getattr__(self, attr):
        if attr in ['Noise', 'AbstractNoise']:
            attr_file = NoiseLoader.class_to_file(attr)
            module = __import__(f'olympus.noises.{attr_file}', fromlist=[attr])
            _class = getattr(module, attr)
            setattr(self, attr, _class)
            return _class
        else:
            noise = NoiseLoader.import_noise(attr)
            setattr(self, attr, noise)
            return noise

    @staticmethod
    def class_to_file(class_name):
        file_name = class_name[0].lower()
        for character in class_name[1:]:
            if character.isupper():
                file_name += f'_{character.lower()}'
            else:
                file_name += character
        return file_name

    @staticmethod
    def file_to_class(file_name):
        class_name = file_name[0].upper()
        next_upper = False
        for character in file_name[1:]:
            if character == '_':
                next_upper = True
                continue
            if next_upper:
                class_name += character.upper()
            else:
                class_name += character
            next_upper = False
        return class_name

    @staticmethod
    def import_noise(attr):
        attr_file = NoiseLoader.class_to_file(attr)
        module = __import__(f'olympus.noises.noise_{attr_file}', fromlist=[attr])
        _class = getattr(module, attr)
        return _class

    def _find_noise(self):
        self.noise_files = []
        self.noise_names = []
        self.noise_map  = {}
        for dir_name in glob.glob(f'{__home__}/noise_*'):
            noise_name = dir_name.split('/')[-1][6:]
            self.noise_files.append(noise_name)
            self.noise_names.append(NoiseLoader.file_to_class(noise_name))

    def get_noises_list(self):
        return sorted(self.noise_names)

    def list_noises(self):
        return sorted(self.noise_names)


sys.modules[__name__] = NoiseLoader(**locals())




#!/usr/bin/env python

from olympus.utils.misc import check_planner_module

#===============================================================================

planner = 'LatinHypercube'
module  = 'pyDOE'
link    = 'https://pythonhosted.org/pyDOE/'
check_planner_module(planner, module, link)

#===============================================================================

from .wrapper_latin_hypercube import LatinHypercube



#!/usr/bin/env python

import os
import glob
import pickle

from olympus import Object
from olympus import Database
from olympus import Logger
from olympus import list_datasets

__home__ = os.path.dirname(os.path.abspath(__file__))

# ==============================================================================


class Baseline(Object):
    def __init__(self, summary_file=f"{__home__}/baseline_summary.pkl"):
        Object.__init__(**locals())
        self._load_baselines()

    def _load_summaries(self, datasets_):
        datasets = datasets_.copy()
        self.baseline_summaries = {}
        # nothing to do if the summary is not available
        if not os.path.isfile(self.summary_file):
            return
        # load summary
        with open(self.summary_file, "rb") as content:
            baseline_summaries = pickle.load(content)
        for dataset, summary in baseline_summaries.items():
            if dataset in datasets:
                self.baseline_summaries[dataset] = summary
                datasets.remove(dataset)
            else:
                Logger.log(
                    f"found summary for not reported dataset: {dataset}", "WARNING"
                )
        for dataset in datasets:
            Logger.log(f"could not find summary for dataset: {dataset}", "WARNING")

    def _register_dbs(self, datasets_):
        # only register complete baselines
        datasets = datasets_.copy()
        self.baseline_db_files = {}
        self.baseline_dbs = {}
        for db_file in glob.glob(f"{__home__}/db_baseline_*sqlite"):
            dataset = db_file.split("/")[-1].split(".")[0][12:]
            self.baseline_db_files[dataset] = db_file
            if dataset in datasets:
                datasets.remove(dataset)
            else:
                Logger.log(
                    f"found complete baseline for not reported dataset: {dataset}",
                    "WARNING",
                )
        for dataset in datasets:
            Logger.log(
                f"could not find complete baseline for dataset: {dataset}", "WARNING"
            )

    def _load_baseline_db(self, dataset):
        file_name = self.baseline_db_files[dataset]
        db = Database().from_file(file_name)
        self.baseline_dbs[dataset] = db
        return db

    def _load_baselines(self):
        # get list of datasets for which we expect baseline
        datasets = list_datasets()
        # load summaries of baseline
        self._load_summaries(datasets)
        # get file names and paths for complete baselines
        self._register_dbs(datasets)

    def get_summary(self, dataset):
        if dataset in self.baseline_summaries.keys():
            return self.baseline_summaries[dataset]
        else:
            Logger.log(f"could not find summary for dataset: {dataset}", "ERROR")

    def get_db(self, dataset):
        if dataset in self.baseline_dbs.keys():
            return self.baseline_dbs[dataset]
        elif dataset in self.baseline_db_files.keys():
            self._load_baseline_db(dataset)
            return self.baseline_dbs[dataset]
        else:
            Logger.log(f"could not find baseline db for dataset: {dataset}", "ERROR")

    def get_campaigns(self, dataset):
        if dataset in self.baseline_dbs.keys():
            return [campaign for campaign in self.baseline_dbs[dataset]]
        elif dataset in self.baseline_db_files.keys():
            self._load_baseline_db(dataset)
            return [campaign for campaign in self.baseline_dbs[dataset]]
        else:
            Logger.log(f"could not find baseline db for dataset: {dataset}", "ERROR")

    def get(self, dataset, kind="summary"):
        """ Retrieves baseline for a given dataset

        Args:
            dataset (str): name of the dataset for which baseline should be retrieved
            kind (str): indicates format of baseline; choose from "summary", "db", or "campaigns"

        Returns:
            requested baseline
        """
        if kind == "summary":
            return self.get_summary(dataset)
        elif kind == "db":
            return self.get_db(dataset)
        elif kind == "campaigns":
            return self.get_campaigns(dataset)
        else:
            Logger.log(
                f'could not understand kind: "{kind}". Please choose from "summary", "db", or "campaigns"',
                "ERROR",
            )



#!/usr/bin/env python

import time
import numpy as np 

from olympus.objects                   import ParameterVector
from olympus.planners.abstract_planner import AbstractPlanner


#===============================================================================

class RandomSearch(AbstractPlanner):    

    def __init__(self, goal='minimize', seed=None):
        AbstractPlanner.__init__(**locals())
        if self.seed is not None:
            np.random.seed(seed)

    def _set_param_space(self, param_space):
        self.param_space = param_space

    def _tell(self, observations):
        # random search does not need observations
        pass

    def _ask(self):
        vals = []
        for param in self.param_space:
            if param.type == 'continuous': 
                val = np.random.uniform(low=param.low, high=param.high)
            else:
                raise NotImplementedError
            vals.append(val)
        vals = np.array(vals)
        return ParameterVector().from_array(vals, self.param_space)






#!/usr/bin/env python

from hyperopt import fmin, tpe, hp, Trials, STATUS_OK, JOB_STATE_DONE
from collections import OrderedDict

from olympus.planners import AbstractPlanner
from olympus.objects import ParameterVector


class Hyperopt(AbstractPlanner):

    def __init__(self, goal='minimize', show_progressbar=False):
        """
        Tree of Parzen Estimators (TPE) as implemented in HyperOpt.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            show_progressbar (bool): If True, show a progressbar.
        """
        AbstractPlanner.__init__(**locals())
        self._trials   = Trials()  # these is a Hyperopt object that stores the search history
        self._hp_space = None  # these are the params in the Hyperopt format

    def _set_param_space(self, param_space):
        self._param_space = []
        for param in param_space:
            if param.type == 'continuous':
                param_dict = {'name': param.name, 'type': param.type, 'domain': (param.low, param.high)}
            self._param_space.append(param_dict)
        # update hyperopt space accordingly
        self._set_hp_space()

    def _tell(self, observations):
        self._params = observations.get_params(as_array=False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
        # update hyperopt Trials accordingly
        self._set_hp_trials()

    def _set_hp_space(self):
        space = []
        # go through all parameters we have defined and convert them to Hyperopt format
        for param in self._param_space:
            if param['type'] == 'continuous':
                space.append((param['name'], hp.uniform(param['name'], param['domain'][0], param['domain'][1])))
        # update instance attribute that is the space input for Hyperopt fmin
        self._hp_space = OrderedDict(space)

    def _set_hp_trials(self):
        self._trials = Trials()
        if self._params is not None and len(self._params) > 0:
            for tid, (param, loss) in enumerate(zip(self._params, self._values)):
                idxs = {k: [tid] for k, v in param.items()}
                vals = {k: [v] for k, v in param.items()}
                hyperopt_trial = Trials().new_trial_docs(
                                        tids=[tid],
                                        specs=[None],
                                        results=[{'loss': loss, 'status': STATUS_OK}],
                                        miscs=[{'tid': tid,
                                                'cmd': ('domain_attachment', 'FMinIter_Domain'),
                                                'idxs': idxs,
                                                'vals': vals,
                                                'workdir': None}]
                                        )
                hyperopt_trial[0]['state'] = JOB_STATE_DONE
                self._trials.insert_trial_docs(hyperopt_trial)
                self._trials.refresh()

    def _ask(self):
        # NOTE: we pass a dummy function as we just ask for the new (+1) set of parameters
        _ = fmin(fn=lambda x: 0, space=self._hp_space, algo=tpe.suggest, max_evals=self.num_generated,
                 trials=self._trials, show_progressbar=self.show_progressbar)

        # make sure the number of parameters asked matches the number of Hyperopt iterations/trials
        assert len(self._trials.trials) == self.num_generated
        # get params from last dict in trials.trials
        proposed_params = self._trials.trials[-1]['misc']['vals']
        for key, value in proposed_params.items():
            proposed_params[key] = value[0]  # this is just to make value not a list

        return ParameterVector(dict=proposed_params, param_space=self.param_space)



#!/usr/bin/env python

import time
import numpy as np 

from olympus                           import Logger
from olympus.objects                   import ParameterVector
from olympus.planners.abstract_planner import AbstractPlanner

#===============================================================================

class LatinHypercube(AbstractPlanner):    

    def __init__(self, goal='minimize', budget=None):
        AbstractPlanner.__init__(**locals())
        self.has_optimizer = False
        self.budget = budget

    def _set_param_space(self, param_space):
        self.param_space = param_space
        self.dim = len(self.param_space)

    def _tell(self, observations):
        self._params = observations.get_params(as_array=False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)

    def _create_optimizer(self):
        from pyDOE import lhs 
        if self.budget is None:
            message = (f'Please provide a number of samples for this planner. Given no number of samples provided, '
                       f'falling back to setting `budget` to {len(self.param_space)}')
            Logger.log(message, 'WARNING')
            self.budget = len(self.param_space)
        self.samples = lhs(len(self.param_space), samples=self.budget)
        for index, param in enumerate(self.param_space):
            self.samples[:, index] = (param.high - param.low) * self.samples[:, index] + param.low 
        self.samples = list(self.samples)
        self.has_optimizer = True

    def _ask(self):
        if not self.has_optimizer:
            self._create_optimizer()
        param = self.samples.pop(0)
        return ParameterVector().from_array(param, self.param_space)







#!/usr/bin/env python

import os
import subprocess

from olympus import __home__
from olympus import Logger


class ConnectorGithub:

    URL = "https://github.com/FlorianHase/olympus_datasets/branches/datasets/src/olympus_datasets/"

    def __init__(self):
        pass

    def list(self):
        Logger.log("connecting to github", "INFO")
        tmp_file = "remote_folders"
        remote_datasets = []
        subprocess.call(f"svn ls -R {self.URL} > {tmp_file}", shell=True)
        with open(tmp_file, "r") as content:
            for line in content:
                dataset_name = line.split("/")[0]
                if not dataset_name in remote_datasets:
                    remote_datasets.append(dataset_name)
        os.remove(tmp_file)
        remote_datasets = [remote_dataset[8:] for remote_dataset in remote_datasets]
        return sorted(remote_datasets)

    def get_dataset(self, dataset_name):
        url = f"{self.URL}/datasets/dataset_{dataset_name}"
        subprocess.call(
            f"svn export {url} {__home__}/datasets/dataset_{dataset_name}", shell=True
        )
        # check if expected files exist
        expected_files = ["dataset.zip", "data.csv", "description.txt"]
        success = True
        for expected_file in expected_files:
            success = success and os.path.isfile(
                f"{__home__}/datasets/dataset_{dataset_name}/{expected_file}"
            )
        return success

    def get_baseline(self):
        url = f"{self.URL}/baseline/"
        subprocess.call(
            f"svn export {url} {__home__}/baseline/random_baseline", shell=True
        )
        success = True
        return success



#!/usr/bin/env python

from olympus import Logger

from . import get_planners_list
from . import import_planner
from . import AbstractPlanner


# NOTE: This goes against the python convention of having function names being
# lowercase, but I thought Planner still works in this case as it effectively
# returns a class instance
# A function seems to me the easiest way to do this right now, but another
# option would be to have a class and fiddle with __new__ an return an instance
# of a different class. I would stick with the simplest option that achieve
# what we need though, unless we already foresee possible expansions that need a
# more complex object
def Planner(kind="ConjugateGradient", goal="minimize", param_space=None):
    """Convenience function to access planners via a slightly higher level interface It returns a certain planner
    with defaults arguments by keyword.

    Args:
        kind (str or AbstractPlanner): Keyword identifying one of the algorithms available in Olympus. Alternatively,
            you can pass a custom algorithm that is a subclass of AbstractPlanner.
        goal (bool): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
        param_space (ParamSpace): A ParameterSpace object defining the space over which to search.

    Returns:
        Planner: An instance of the chosen planning algorithm.
    """
    _validate_planner_kind(kind)
    # if a string is passed, then load the corresponding wrapper
    if type(kind) == str:
        from . import PlannerLoader

        kind = PlannerLoader.file_to_class(kind)
        planner = import_planner(kind)
        planner = planner(goal=goal)
    # if an instance of a planner is passed, simply return the same instance
    elif isinstance(kind, AbstractPlanner):
        planner = kind
    # if a custom class is passed, then that is the 'wrapper'
    elif issubclass(kind, AbstractPlanner):
        planner = kind()

    # load param_space already if provided, otherwise it will have to be set by self.set_param_space
    if param_space is not None:
        planner.set_param_space(param_space)

    return planner


def _validate_planner_kind(kind):
    # if we received a string
    if type(kind) == str:
        from . import PlannerLoader

        kind = PlannerLoader.file_to_class(kind)
        avail_planners = get_planners_list()
        if kind not in avail_planners:
            message = (
                'Planner "{0}" not available in Olympus. Please choose '
                "from one of the available planners: {1}".format(
                    kind, ", ".join(avail_planners)
                )
            )
            Logger.log(message, "FATAL")

    # if we get an instance of a planner class
    elif isinstance(kind, AbstractPlanner):
        # make sure it has the necessary methods
        for method in ["_set_param_space", "_tell", "_ask"]:
            implementation = getattr(kind, method, None)
            if not callable(implementation):
                message = f'The object {kind} does not implement the necessary method "{method}"'
                Logger.log(message, "FATAL")

    # if we received a custom planner class
    elif issubclass(kind, AbstractPlanner):
        # make sure it has the necessary methods
        for method in ["_set_param_space", "_tell", "_ask"]:
            implementation = getattr(kind, method, None)
            if not callable(implementation):
                message = f'The object {kind} does not implement the necessary method "{method}"'
                Logger.log(message, "FATAL")

    # if we do not know what was passed raise an error
    else:
        message = 'Could not initialize Planner: the argument "kind" is neither a string or AbstractPlanner subclass'
        Logger.log(message, "FATAL")



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface
from scipy.stats import multivariate_normal
from olympus import Logger


class GaussianMixture(AbstractSurface):

    def __init__(self, param_dim=2, num_gauss=2, cov_scale=0.5, diagonal_cov=False, means=None, covariances=None,
                 noise=None, random_seed=None):
        """Random Gaussian Mixture surface. Given the number of dimensions and Gaussian distributions provided,
        the surface is the sum of probability densities. Unless a set of means and covariances are provided,
        a set of random means and covariances are generated.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            num_gauss (int): Number of Gaussians. Default is 2.
            cov_scale (float): Scaling factor for the randomly generated covariance matrices (default is 0.5).
                A random covariance matrix B is generated by multiplying a random matrix A by its transpose.
                The elements of the random matrix are drawn from a uniform distribution. `cov_scale` is used to scale
                the resulting covariance matrix. Setting it to one typically generates smooth surfaces, while setting
                it to 0.1 generates rougher surfaces.
            diagonal_cov (bool): Whether to use a diagonal rather than full covariance matrix. Default is False.
            means (array): 2-dimensional array (shape=(#gaussians, #dimensions)) providing a list of means for the
                Gaussians. If provided, `param_dim` and `num_gauss` will be ignored, as they will be derived from the
                dimensionality of this array.
            covariances (array): 3-dimensional array (shape=(#gaussians, #dimensions, #dimensions)) providing a list
                of covariance matrices for the Gaussians. If provided, `param_dim` and `num_gauss` will be ignored,
                as they will be derived from the dimensionality of this array.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
            random_seed (int): Fix the random seed for reproducible surfaces. Default is None, i.e. random.
        """
        param_dim, _num_gauss = self._determine_param_dim_and_num_gauss(param_dim, num_gauss, means, covariances)
        AbstractSurface.__init__(**locals())
        np.random.seed(self.random_seed)  # set random seed
        self._create_gaussian_mixture()

    @staticmethod
    def _determine_param_dim_and_num_gauss(param_dim, num_gauss, means, covariances):

        # no means and covs provided
        if means is None and covariances is None:
            _num_gauss = num_gauss
            _param_dim = param_dim
        # means provided
        elif means is not None and covariances is None:
            _num_gauss = np.shape(means)[0]
            _param_dim = np.shape(means)[1]
        # covs provided
        elif means is not None and covariances is not None:
            _num_gauss = np.shape(covariances)[0]
            _param_dim = np.shape(covariances)[1]
        # both means and covs provided
        elif means is not None and covariances is not None:
            assert np.shape(means)[0] == np.shape(covariances)[0]
            assert np.shape(means)[1] == np.shape(covariances)[1]
            _num_gauss = np.shape(means)[0]
            _param_dim = np.shape(means)[1]

        return _param_dim, _num_gauss

    @property
    def minima(self):
        # needs to be found by optimization
        message = 'Unknown minima: these need to be found numerically'
        Logger.log(message, 'WARNING')
        return None

    @property
    def maxima(self):
        # needs to be found by optimization
        message = 'Unknown maxima: these need to be found numerically'
        Logger.log(message, 'WARNING')
        return None

    def _run(self, params):
        value = - np.sum([g.pdf(params) for g in self.gaussians], axis=0)
        if self.noise is None:
            return value
        else:
            return self.noise(value)

    def _create_gaussian_mixture(self):

        # -------------------------
        # Get means and covariances
        # -------------------------
        # if means are provided
        if self.means is not None:
            self._means = np.array(self.means)
        # else generate random ones
        else:
            self._generate_random_means()

        # if covariances provided
        if self.covariances is not None:
            self._covs = np.array(self.covariances)
        # else generate random ones
        else:
            self._generate_random_covs()

        # ----------------
        # Create Gaussians
        # ----------------
        self.gaussians = []
        for i in range(self._num_gauss):
            g = multivariate_normal(mean=self._means[i], cov=self._covs[i])
            self.gaussians.append(g)

    def _generate_random_means(self):
        self._means = []
        for i in range(self._num_gauss):
            mean = np.random.uniform(low=0, high=1, size=len(self.param_space))
            self._means.append(mean)

    def _generate_random_covs(self):
        """

        Args:
            size:
            scale (float): scale the random covariance matrix.

        Returns:

        """
        self._covs = []

        if self.diagonal_cov is False:
            for i in range(self._num_gauss):
                # generate random positive semi-definite matrix, scale by sigma_scale
                A = np.random.rand(len(self.param_space), len(self.param_space))
                alpha = 0.01
                cov = np.dot(A, A.transpose()) + alpha * np.diag(np.ones(len(self.param_space)))
                self._covs.append(cov * self.cov_scale)
        elif self.diagonal_cov is True:
            for i in range(self._num_gauss):
                diag = np.random.uniform(low=0, high=1, size=len(self.param_space)) * self.cov_scale
                cov = np.diag(np.array(diag))
                self._covs.append(cov)




#!/usr/bin/env python

import numpy as np
from olympus import Logger
from abc import abstractmethod
from olympus.objects import Object, abstract_attribute, ABCMeta, ParameterVector
from olympus.campaigns.param_space import ParameterSpace, ParameterContinuous


class AbstractNoise(Object, metaclass=ABCMeta):

    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)

    @abstract_attribute
    def scale(self):
        pass

    @abstractmethod
    def _add_noise(self, value):
        pass

    @property
    def kind(self):
        return type(self).__name__

    # TODO: possible extension is to take also as argument params, to allow for noise that depends on the
    #  location in param space, and not just output location
    def __call__(self, value):
        """Add noise to the provided values.

        Args:
            values (float): Measurement to which noise is to be added.

        Returns:
            noisy_values (array): Noisy measurement.
        """
        return self._add_noise(value)




#!/usr/bin/env python

# ==============================================================================

from olympus import Logger

from .connector_server import ConnectorServer
from .connector_github import ConnectorGithub

# ==============================================================================


class ParserBaseline:
    def __init__(self, subparsers):
        self.parser = subparsers.add_parser("baseline", help=">> help for baseline")
        self.parser.add_argument("--get", dest="get", action="store_true")
        self.parser.set_defaults(get=False)

    @staticmethod
    def __call__(args):
        get = args.get

        if get is True:

            for Connector in [ConnectorGithub, ConnectorServer]:
                success = Connector().get_baseline()
                if success:
                    break

        elif get is False:
            pass



#!/usr/bin/env python

import numpy as np
from abc import abstractmethod
from olympus import Logger
from olympus.campaigns import Observations, Campaign
from olympus.objects import Object, abstract_attribute, ABCMeta, Config


class AbstractPlanner(Object, metaclass=ABCMeta):
    """ This class is intended to contain methods shared by all wrappers, as well as define abstract methods and
    attributes that all planner wrappers need to implement. It is not meant to be exposed to the standard user,
    although more advanced users can use this class to create custom planner classes.
    """

    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)
        self.num_generated = 0
        self.param_space = None
        self._params = None
        self._values = None
        self.SUBMITTED_PARAMS = []
        self.RECEIVED_VALUES  = []

        # rm all those vars in config that are not needed/used by ScipyWrapper
        for var in ['goal', 'init_guess', 'random_seed']:
            if var in kwargs:
                del kwargs[var]
        self.config = Config(from_dict=kwargs)

        # self.goal is an abstract attribute that needs to be defined by the subclasses of AbstractPlanner
        # Since all planner wrappers are implemented in minimization mode, we flip the measurements if we want to
        # perform a maximization
        if self.goal == 'minimize':
            self.flip_measurements = False
        elif self.goal == 'maximize':
            self.flip_measurements = True
        else:
            message = f'attribute `goal` can only be "minimize" or "maximize". "{self.goal}" is not a valid value'
            Logger.log(message, 'ERROR')

    def __repr__(self):
        if not hasattr(self, 'param_space') or self.param_space is None:
            param_space = 'undefined'
        else:
            param_space = 'defined'
        return f"<Planner (kind={self.kind}, param_space={param_space})>"

    @abstract_attribute
    def goal(self):
        pass

    @property
    def kind(self):
        return type(self).__name__

    @abstractmethod
    def _set_param_space(self, *args, **kwargs):
        """Method that returns the parameter space over which to optimize in the format used by a specific optimization
        library/wrapper.
        """
        pass

    @abstractmethod
    def _tell(self, *args, **kwargs):
        """Method that returns all previous observations as required by a specific optimization library/wrapper.
        """
        pass

    @abstractmethod
    def _ask(self, *args, **kwargs):
        """Method that returns the next query point based on all previous observations.
        """
        pass

    def set_param_space(self, param_space):
        """Defines the parameter space over which the planner will search.

        Args:
            param_space (ParameterSpace): a ParameterSpace object defining the space over which to search.
        """
        self.param_space = param_space
        self._set_param_space(param_space)

    def ask(self, return_as=None):
        """ suggest new set of parameters

        Args:
            return_as (string): choose data type for returned parameters
                allowed options (dict, array)

        Returns:
            ParameterVector: newly generated parameters
        """

        self.num_generated += 1
        param_vector = self._ask()

        # check that the parameters suggested are within the bounds of our param_space
        self._validate_paramvector(param_vector)

        if return_as is not None:
            try:
                param_vector = getattr(param_vector, 'to_{}'.format(return_as))()
            except AttributeError:
                Logger.log('could not return param_vector as "{}"'.format(return_as), 'ERROR')
        return param_vector

    def tell(self, observations=Observations()):
        """Provide the planner with all previous observations.

        Args:
            observations (Observations): an Observation object containing all previous observations. This defines the
                history of the campaign seen by the planner. The default is None, i.e. there are no previous
                observations.
        """
        self._tell(observations)

    def recommend(self, observations=None, return_as=None):
        """Consecutively executes tell and ask: tell the planner about all previous observations, and ask about the
        next query point.

        Args:
            observations (list of ???)
            return_as (string): choose data type for returned parameters
                allowed options (dict, array)

        Returns:
            list: newly generated parameters
        """
        self.tell(observations)
        return self.ask(return_as=return_as)

    def optimize(self, emulator, num_iter=1, verbose=False):
        """Optimizes a surface for a fixed number of iterations.

        Args:
            emulator (object): Emulator or a Surface instance to optimize over.
            num_iter (int): Maximum number of iterations allowed.
            verbose (bool): Whether to print information to screen.

        Returns:
            campaign (Campaign): Campaign object with information about the optimization, including all parameters
                tested and measurements obtained.
        """

        # update num_iter if needed by the specific wrapper
        if hasattr(self, 'num_iter') and self.num_iter != num_iter:
            Logger.log(f'Updating the number of sampling points of planner {type(self).__name__} to {num_iter}', 'INFO')
            self.num_iter = num_iter

        # same for budget
        if hasattr(self, 'budget') and self.budget != num_iter:
            Logger.log(f'Updating the number of sampling points of planner {type(self).__name__} to {num_iter}', 'INFO')
            self.budget = num_iter

        # reset planner if it has a 'reset' method. Assuming that if there is a 'reset' method it is needed here
        # This is used by Deap for example, to clear the latest population before doing another optimization
        if callable(getattr(self, "reset", None)):
            self.reset()

        # use campaign to store info, and then to be returned
        campaign = Campaign()
        campaign.set_planner_specs(self)
        campaign.set_emulator_specs(emulator)

        # provide the planner with the parameter space.
        # param space in emulator as it originates from dataset
        self.set_param_space(emulator.param_space)

        # Optimize: i.e. call the planner recommend method for max_iter times
        for i in range(num_iter):
            if verbose is True:
                Logger.log(f'Optimize iteration {i+1}', 'INFO')
                Logger.log(f'Obtaining parameters from planner...', 'INFO')
            # get new params from planner
            # NOTE: now we get 1 param at a time, a possible future expansion is to return batches
            params = self.recommend(observations=campaign.observations)

            # get measurement from emulator/surface
            if verbose is True:
                Logger.log(f'Obtaining measurement from emulator...', 'INFO')
            values = emulator.run(params.to_array(), return_paramvector=True)

            # store parameter and measurement pair in campaign
            campaign.add_observation(params, values)

        return campaign

    def _validate_paramvector(self, param_vector):
        for key, value in param_vector.to_dict().items():
            param = self.param_space.get_param(name=key)
            if param['type'] == 'continuous':
                if not param['low'] <= value <= param['high']:
                    message = 'Proposed parameter {0} not within defined bounds ({1},{2})'.format(value, param['low'], param['high'])
                    Logger.log(message, 'WARNING')

    def _project_into_domain(self, params):
        for _, bound in enumerate(self.param_space):
            params[_] = np.maximum(bound['low'], params[_])
            params[_] = np.minimum(bound['high'], params[_])
        return params



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface
from itertools import product


class Branin(AbstractSurface):

    def __init__(self, noise=None):
        """Branin function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        param_dim = 2
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        # 3 global minima
        x0 = (np.array([-np.pi, np.pi, 9.42478]) + 5) / 15  # rescale onto unit square
        x1 = np.array([12.275, 2.275, 2.475]) / 15  # rescale onto unit square
        params0 = np.array([x0[0], x1[0]])
        params1 = np.array([x0[1], x1[1]])
        params2 = np.array([x0[2], x1[2]])
        value0 = self._run(params0)
        value1 = self._run(params1)
        value2 = self._run(params2)

        # check the minima are the same
        np.testing.assert_almost_equal(value0, value1)
        np.testing.assert_almost_equal(value0, value2)

        return [{'params': params0, 'value': value0},
                {'params': params1, 'value': value1},
                {'params': params2, 'value': value2}]

    @property
    def maxima(self):
        params = [0., 0.]
        value = self._run(params)
        return {'params': params, 'value': value}

    def _run(self, params):
        params = np.array(params)
        # we always only receive 2-dimensional arrays
        x0 = 15 * params[0] - 5  # rescale onto [-5, 10]
        x1 = 15 * params[1]  # rescale onto [0, 15]

        # Branin params
        a = 1
        b = 5.1 / (4*np.pi**2)
        c = 5/np.pi
        r = 6
        s = 10
        t = 1/(8*np.pi)

        result = a * (x1 - b*(x0**2) + c*x0 - r) ** 2 + s * (1 - t) * np.cos(x0) + s

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python
import argparse
from olympus.cli.main import check_unknown_cmd


# ======================
# Parse Option Arguments
# ======================
def parse_options():

	parser = argparse.ArgumentParser(description='It does this and that.')

	parser.add_argument('-g',
						metavar='generator',
						dest='generatot',
						type=str,
						help='the generator to use',
						default='gpyopt')
	parser.add_argument('-e',
						metavar='evaluator',
						dest='evaluator',
						type=str,
						help='the evaluator to use',
						default='dummy')

	args, unknown = parser.parse_known_args()
	check_unknown_cmd(unknown)

	return args


# ====
# Main
# ====
def main(args):
	pass


def entry_point():
	args = parse_options()
	main(args)


if __name__ == '__main__':
	entry_point()



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface
from itertools import product
from olympus import Logger


class DiscreteAckley(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """Discrete Ackley function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        # minimum at the centre
        params = [0.5] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        message = 'DiscreteAckley has an infinite number of maxima'
        Logger.log(message, 'INFO')
        # some maxima
        maxima = []
        params = product([0.05, 0.95], repeat=self.param_dim)
        for param in params:
            param = list(param)
            value = self._run(param)
            maxima.append({'params': param, 'value': value})
        return maxima

    def _run(self, params):
        params = np.array(params)
        params = 100 * params - 50
        bounds = [1.0, 2.0, 3.0, 4.0, 5.0]
        bounds = np.array(bounds) ** 2
        result = 5
        for bound in bounds[::-1]:
            if np.amax(np.abs(params)) < bound:
                result -= 1
        bounds = [1.25, 2.0, 2.5]
        bounds = np.array(bounds) ** 2

        domain = np.linspace(-50, 50, 10)
        dx = domain[1] - domain[0]
        imaged = np.array([np.amin(np.abs(element - domain)) for element in params])
        new_res = 5
        for bound in bounds[::-1]:
            if np.amax(np.abs(imaged)) < bound:
                new_res -= 1
            result = np.amin([result, new_res])
        result = np.amin([4, result])

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

from olympus.planners.wrapper_scipy import WrapperScipy
from olympus.objects.object_config  import Config


class Lbfgs(WrapperScipy):

    METHOD       = 'L-BFGS-B'
    KNOWS_BOUNDS = True

    # defaults are copied from scipy documentation
    # --> https://docs.scipy.org/doc/scipy/reference/optimize.minimize-lbfgsb.html#optimize-minimize-lbfgsb
    def __init__(self, goal='minimize', disp=None, eps=1e-8, ftol=2.220446049250313e-9, gtol=1e-5,
                 maxcor=10, maxfun=15000, maxiter=15000, maxls=20,
                 init_guess=None, init_guess_method='random', init_guess_seed=None):
        """
        L-BFGS-B optimizer based on the SciPy implementation.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            disp (None or int): If disp is None (the default), then the supplied
                version of iprint is used. If disp is not None, then it
                overrides the supplied version of iprint with the behaviour you
                outlined.
            eps (float): Step size used for numerical approximation of the
                jacobian.
            ftol (float): The iteration stops when
                (f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= ftol.
            gtol (float): The iteration will stop when
                max{|proj g_i | i = 1, ..., n} <= gtol where pg_i is the i-th
                component of the projected gradient.
            maxcor (int): The maximum number of variable metric corrections used
                to define the limited memory matrix. (The limited memory BFGS
                method does not store the full hessian but uses this many terms
                in an approximation to it.)
            maxfun (int): Maximum number of function evaluations.
            maxiter (int): Maximum number of iterations.
            maxls (int, optional): Maximum number of line search steps (per
                iteration).
            init_guess (array, optional): initial guess for the optimization
            init_guess_method (str): method to construct initial guesses if init_guess is not provided.
                Choose from: random
            init_guess_seed (str): random seed for init_guess_method
        """

        WrapperScipy.__init__(**locals())



#!/usr/bin/env python

import silence_tensorflow

silence_tensorflow.silence_tensorflow()
import tensorflow as tf
import tensorflow_probability as tfp

from olympus.models import WrapperTensorflowModel

tfd = tfp.distributions

# ===============================================================================


class BayesNeuralNet(WrapperTensorflowModel):

    ATT_KIND = {"type": "string", "default": "BayesNeuralNet"}

    def __init__(
        self,
        scope="model",
        hidden_depth=3,
        hidden_nodes=48,
        hidden_act="leaky_relu",
        out_act="linear",
        learning_rate=1e-3,
        pred_int=100,
        reg=0.001,
        es_patience=100,
        max_epochs=100000,
        batch_size=20,
    ):
        """Bayesian Neural Network model.

        Args:
            scope (str): TenforFlow scope.
            hidden_depth (int): Number of hidden layers.
            hidden_nodes (int): Number of hidden nodes per layer.
            hidden_act (str): Hidden activation function. Available options are 'linear', 'leaky_relu', 'relu',
                'softmax', 'softplus', 'softsign', 'sigmoid'.
            out_act (str): Output activation function. Available options are 'linear', 'leaky_relu', 'relu',
                'softmax', 'softplus', 'softsign', 'sigmoid'.
            learning_rate (float): Learning rate.
            pred_int (int): Frequency with which we make predictions on the validation/training set (in number of epochs).
            reg (float): ???
            es_patience (int): Early stopping patience.
            max_epochs (int): Maximum number of epochs allowed.
            batch_size (int): Size batches used for training.
        """

        WrapperTensorflowModel.__init__(**locals())

    def _build_inference(self):
        self.graph = tf.Graph()
        self.is_graph_constructed = True

        activation_hidden = self.act_funcs(self.hidden_act)
        activation_out = self.act_funcs(self.out_act)

        with self.graph.as_default():

            self.tf_x = tf.compat.v1.placeholder(
                tf.float32, [self.batch_size, self.features_dim]
            )
            self.tf_y = tf.compat.v1.placeholder(
                tf.float32, [self.batch_size, self.targets_dim]
            )

            with tf.name_scope(self.scope):
                self.layers = [
                    tfp.layers.DenseLocalReparameterization(
                        self.hidden_nodes, activation=activation_hidden
                    )
                    for _ in range(self.hidden_depth)
                ]
                self.layers.append(
                    tfp.layers.DenseLocalReparameterization(
                        self.targets_dim, activation=activation_out
                    )
                )

                self.neural_net = tf.keras.Sequential(self.layers)

                self.y_pred = self.neural_net(self.tf_x)
                self.scale = tf.nn.softplus(
                    tf.Variable(tf.ones(self.y_pred.get_shape()))
                )
                self.y_sample = tfd.Normal(loc=self.y_pred, scale=self.scale)

            self.kl = 0
            for _, layer in enumerate(self.layers):
                self.kl += sum(layer.losses) / float(self.batch_size)
            self.reg_loss = -tf.reduce_sum(self.y_sample.log_prob(self.tf_y))
            self.loss = self.reg_loss + (self.reg * self.kl)

            self.optimizer = tf.compat.v1.train.AdamOptimizer(
                learning_rate=self.learning_rate
            )
            self.train_op = self.optimizer.minimize(self.loss)

            self.init_op = tf.group(
                tf.compat.v1.global_variables_initializer(),
                tf.compat.v1.local_variables_initializer(),
            )
            self.sess = tf.compat.v1.Session(graph=self.graph)
            with self.sess.as_default():
                self.sess.run(self.init_op)
                self.saver = tf.compat.v1.train.Saver()



#!/usr/bin/env python

import numpy as np
from olympus.noises import AbstractNoise


class UniformNoise(AbstractNoise):

    def __init__(self, scale=1):
        """Gaussian noise.

        Args:
            scale (float): Half-range :math:`r_{1/2}` of the uniform distribution. A random sample will be drawn from
                the half-open interval :math:`[\mu - r_{1/2}, \mu + r_{1/2})`, where :math:`\mu` the argument ``value``. Default is 1.
        """
        AbstractNoise.__init__(**locals())

    def _add_noise(self, value):
        low = value - self.scale * 0.5
        high = value + self.scale * 0.5
        noisy_value = np.random.uniform(low=low, high=high)
        return noisy_value



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface
from itertools import product


class Schwefel(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """Schwefel function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        # minimum at 420.9687
        min_loc = (420.9687 + 500) / 1000  # rescale onto unit hypercube
        params = [min_loc] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        # is there an analytical maximum?
        return None

    def _run(self, params):
        params = 1000 * np.array(params) - 500  # rescale onto [-500, 500]
        result = 0
        for index, element in enumerate(params):
            result += - element * np.sin(np.sqrt(np.abs(element)))

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

import json
from pandas import DataFrame, read_csv
import numpy as np

from olympus import Logger
from olympus.campaigns import ParameterSpace
from olympus.objects import Parameter

import os
from glob import glob


# =========================
# Main Class of This Module
# =========================
class Dataset:
    """
    A ``Dataset`` object stores the data of a dataset by wrapping a ``pandas.DataFrame`` in its ``data`` attribute, provides
    additional information on the dataset, and provides convenience methods to access features and targets as well as
    to generate training/validation/test splits.

    Args:
        kind (str): kind of the Olympus dataset to load.
        data (array): custom dataset. Same input as for pandas.DataFrame.
        columns (list): column names. Same input as for pandas.DataFrame.
        target_ids (list): list of column indices, or names if provided, that identify the targets for the predictions.
        test_frac (float): fraction of the data to be used as test set.
        num_folds (int): number of cross validation folds the training set will be split into.
        random_seed (int): random seed for numpy. Setting a seed makes the random splits reproducible.
    """

    def __init__(
        self,
        kind=None,
        data=None,
        columns=None,
        target_ids=None,
        test_frac=0.2,
        num_folds=5,
        random_seed=None,
    ):

        _validate_dataset_args(kind, data, columns, target_ids)

        self.kind = kind
        self.name = kind
        self.test_frac = test_frac
        self.num_folds = num_folds
        self.random_seed = random_seed

        # we defined the param_space as part of the Dataset object. When we
        # load a standard dataset, we define the param_space from the config
        # file. If Dataset is a custom dataset, the user will need to set_param_space
        self.param_space = None
        self.value_space = None

        # ------------------------------------------
        # Case 1: data directly provided by the user
        # ------------------------------------------
        if kind is None:
            self._description = "Custom Dataset"
            if target_ids is None:
                self._targets = []
            else:
                self._targets = target_ids
            _data = data

        # ------------------------------
        # Case 2: Olympus dataset loaded
        # ------------------------------
        elif kind is not None:
            _data, _config, self._description = load_dataset(kind)
            self._targets = [t["name"] for t in _config["measurements"]]
            columns = [f["name"] for f in _config["parameters"]] + self._targets
            # create param_space from config file
            self._create_param_space(_config)
            self._create_value_space(_config)
            # define attributes of interest - done here so to avoid calling load_dataset again
            self.constraints = _config["constraints"]
            self._goal = _config["default_goal"]

        # the numeric data is stored here, wrapping a DataFrame
        self.data = DataFrame(data=_data, index=None, columns=columns)

        # make sure the targets are the last column(s)
        for tgt in self._targets:
            assert tgt in list(self.data.columns)[-len(self._targets) :]

        # create dataset splits
        self.create_train_validate_test_splits()

    @property
    def goal(self):
        if not "_goal" in self.__dict__:
            _data, _config, _description = load_dataset(self.kind)
            self._goal = _config["default_goal"]
        return self._goal

    @property
    def measurement_name(self):
        if not "_measurement_name" in self.__dict__:
            _data, _config, _description = load_dataset(self.kind)
            self._measurement_name = _config["measurements"][0]["name"]
        return self._measurement_name

    def __getattr__(self, attr):
        if attr == "__getstate__":
            return lambda: self.__dict__
        elif attr == "__setstate__":

            def set_state(_dict):
                self.__dict__ = _dict

            return set_state
        elif attr == "kind" and not attr in self.__dict__:
            val = getattr(self, "name")
            setattr(self, attr, val)
            return val
        print('ATTR', attr)
        return getattr(self, attr)

    def __str__(self):
        # TODO: add some more info here...
        string = f"Dataset(kind={self.kind})"
        return string

    @property
    def size(self):
        return self.data.shape[0]

    @property
    def shape(self):
        return self.data.shape

    # we could also overwrite the DataFrame.info method
    def dataset_info(self):
        """Provide summary info about dataset.
        """
        Logger.log(self._description, "INFO")

    def set_param_space(self, param_space):
        """Define the parameter space of the dataset.

        Args:
            param_space (ParameterSpace): ParameterSpace object with information about all variables in the dataset.

        """

        # validate first...
        for i, param in enumerate(param_space.parameters):
            # check names are matching
            if self.feature_names[i] != param.name:
                message = f"Parameter name `{self.feature_names[i]}` does not match name `{param.name}` found in dataset!"
                Logger.log(message, "WARNING")
            # check data provided is within param_space bounds
            if param.low > np.min(self.data.iloc[:, i]):
                message = f"Lower bound of {param.low} provided for parameter `{param.name}` is higher than minimum found in the data!"
                Logger.log(message, "ERROR")
            if param.high < np.min(self.data.iloc[:, i]):
                message = f"Upper bound of {param.high} provided for parameter `{param.name}` is lower than maximum found in the data!"
                Logger.log(message, "ERROR")

        # ...then assign
        self.param_space = param_space

    def infer_param_space(self):
        """Guess the parameter space from the dataset. The range for all parameters will be define based on the
        minimum and maximum values in the dataset for each variable. All variables will be assumed not to be periodic.
        """
        param_space = ParameterSpace()
        for name in self.feature_names:
            param_dict = {
                "name": name,
                "low": np.min(self.data.loc[:, name]),
                "high": np.max(self.data.loc[:, name]),
            }
            param = Parameter().from_dict(param_dict)
            param_space.add(param)

        # set the param space
        self.set_param_space(param_space)

    def to_disk(self, folder='custom_dataset'):
        """Save the dataset to disk in the format expected by Olympus for its own datasets. This can be useful if you
        plan to upload the dataset to the community datasets available online.

        Args:
            folder (str): Folder in which to save the dataset files.
        """

        # create the folder
        os.mkdir(folder)

        # save the numeric data
        self.data.to_csv(f'{folder}/data.csv', header=False, index=False)

        # save the description
        self._generate_description()
        with open(f'{folder}/description.txt', 'w') as f:
            f.write(self._description)

        # save the config file
        _config = {}
        _config['constraints'] = {'parameters': 'none', 'measurements': 'none'}

        _config['parameters'] = []
        for param in self.param_space.parameters:
            d = {'name': param.name, 'type': param.kind, 'lower': param.low, 'upper':param.high}
            _config['parameters'].append(d)

        _config['measurements'] = []
        for target in self._targets:
            t = {'name': target, 'type': 'continuous'}
            _config['measurements'].append(t)

        with open(f'{folder}/config.json', 'w') as f:
            f.write(json.dumps(_config, indent=4, sort_keys=True))

    def _generate_description(self):
        _description = []
        if self.kind is None:
            _description.append('Custom Dataset\n')
        else:
            _description.append(f'{self.kind}\n')

        _description.append('=========================================')
        _description.append('                Summary')
        _description.append('-----------------------------------------')
        _description.append(f'    Number of Samples       {self.size:>10}')
        _description.append(f'    Dimensionality          {len(self.feature_names):>10}')
        _description.append(f'    Features:')
        for param in self.param_space.parameters:
            _description.append(f'        {param.name:<10}          {param.kind:>10}')
        _description.append(f'    Targets:')
        for target in self._targets:
            _description.append(f'        {target:<10}          continuous')

        self._description = "\n".join(_description)

    # ----------------------------------
    # Methods about features and targets
    # ----------------------------------
    @property
    def feature_names(self):
        # return names of features
        return [c for c in self.data.columns if c not in self._targets]

    @property
    def target_names(self):
        # return names of targets
        return self._targets

    @property
    def features(self):
        # return data for features
        return self.data.loc[:, self.feature_names]

    @property
    def targets(self):
        # return data for targets
        if type(self.target_names[0]) == str:
            return self.data.loc[:, self.target_names]
        elif type(self.target_names[0]) == int:
            return self.data.iloc[:, self.target_names]

    @property
    def features_dim(self):
        return len(self.feature_names)

    @property
    def targets_dim(self):
        return len(self.target_names)

    # ----------------------------------------------------------------
    # Methods for dataset splitting into training/validation/test sets
    # ----------------------------------------------------------------
    def create_train_validate_test_splits(
        self, test_frac=0.2, num_folds=5, test_indices=None
    ):
        """
        Args:
            test_frac (float)
            num_folds (int)
            test_indices (array)
                Array with the indices of the samples to be used as test set.
        """
        # update num_folds and test frac
        self.test_frac = test_frac
        if test_indices is not None:
            self.test_frac = len(test_indices) / len(self.data)
        self.num_folds = num_folds

        np.random.seed(self.random_seed)  # set random seed
        nrows = self.data.shape[0]
        indices = range(nrows)
        n_train_samples = int(round(nrows * self.test_frac, ndigits=0))

        # test set
        if test_indices is None:
            self.test_indices = np.random.choice(
                indices, size=n_train_samples, replace=False
            )
        else:
            self.test_indices = np.array(test_indices)

        # training set
        self.train_indices = np.setdiff1d(indices, self.test_indices)

        # cross-validation sets are subsets of the training set
        train_indices_shuffled = np.random.permutation(
            self.train_indices
        )  # random shuffle
        self.cv_fold_indices = np.array_split(
            train_indices_shuffled, self.num_folds
        )  # split in approx equal sized arrays
        # define train/valid indices for each fold
        self.cross_val_indices = []
        for i in range(self.num_folds):
            fold_train = np.concatenate(
                np.delete(self.cv_fold_indices, i, axis=0), axis=0
            )
            fold_valid = self.cv_fold_indices[i]
            self.cross_val_indices.append([fold_train, fold_valid])
        # TODO: unfix random seed?

    @property
    def train_set(self):
        # The training set with features+targets
        return self.data.loc[self.train_indices, :]

    @property
    def train_set_features(self):
        # The training set with only the features
        return self.data.loc[self.train_indices, self.feature_names]

    @property
    def train_set_targets(self):
        # The training set with only the targets
        return self.data.loc[self.train_indices, self.target_names]

    @property
    def test_set(self):
        # The test set with features+targets
        return self.data.loc[self.test_indices, :]

    @property
    def test_set_features(self):
        # The test set with only the features
        return self.data.loc[self.test_indices, self.feature_names]

    @property
    def test_set_targets(self):
        # The test set with only the targets
        return self.data.loc[self.test_indices, self.target_names]

    @property
    def cross_val_sets(self):
        # List of cv sets with features+targets
        cv_sets = []
        for train_id, valid_id in self.cross_val_indices:
            train_data = self.data.loc[train_id, :]
            valid_data = self.data.loc[valid_id, :]
            cv_sets.append((train_data, valid_data))
        return cv_sets

    @property
    def cross_val_sets_features(self):
        # List of cv sets with only features
        cv_sets = []
        for train_id, valid_id in self.cross_val_indices:
            train_data = self.data.loc[train_id, self.feature_names]
            valid_data = self.data.loc[valid_id, self.feature_names]
            cv_sets.append((train_data, valid_data))
        return cv_sets

    @property
    def cross_val_sets_targets(self):
        # List of cv sets with only targets
        cv_sets = []
        for train_id, valid_id in self.cross_val_indices:
            train_data = self.data.loc[train_id, self.target_names]
            valid_data = self.data.loc[valid_id, self.target_names]
            cv_sets.append((train_data, valid_data))
        return cv_sets

    def get_cv_fold(self, fold):
        """Get the data for a specific cross-validation fold.

        Args:
            fold (int): fold id.

        Returns:
            data (DataFrame): data for the chosen fold.
        """
        indices = self.cv_fold_indices[fold]
        return self.data.loc[indices, :]

    # ---------------
    # Private Methods
    # ---------------
    def _create_param_space(self, config):
        self.param_space = ParameterSpace()
        self.param_space.add(
            [Parameter().from_dict(feature) for feature in config["parameters"]]
        )

    def _create_value_space(self, config):
        self.value_space = ParameterSpace()
        self.value_space.add(
            [Parameter().from_dict(feature) for feature in config["measurements"]]
        )


# ===============
# Other Functions
# ===============
def load_dataset(kind):
    """Loads a dataset from the Olympus dataset library.

    Args:
        kind (str): kind of Olympus dataset to load.

    Returns:
        (tuple): tuple containing:
            data (array): numpy array where each row is a sample and each column is a feature/target. The columns are
            sorted such features come first and then targets.
            config (dict): dict containing information on the features and targets present in data.
            description (str): string describing the dataset.
    """

    _validate_dataset_args(kind=kind, data=None, columns=None, target_names=None)
    datasets_path = os.path.dirname(os.path.abspath(__file__))

    # load description
    with open("".join(f"{datasets_path}/dataset_{kind}/description.txt")) as txtfile:
        description = txtfile.read()

    # load info on features/targets
    with open("".join(f"{datasets_path}/dataset_{kind}/config.json"), "r") as content:
        config = json.loads(content.read())

    # load data
    csv_file = "".join(f"{datasets_path}/dataset_{kind}/data.csv")
    try:
        data = read_csv(csv_file, header=None).to_numpy()
    except FileNotFoundError:
        Logger.log(f"Could not find data.csv for dataset {kind}", "FATAL")

    return data, config, description


def _validate_dataset_args(kind, data, columns, target_names):
    if kind is not None:
        # -----------------------------------
        # check that a correct name is passed
        # -----------------------------------
        # TODO: reduce redundant code by importing the list from where we have it already
        module_path = os.path.dirname(os.path.abspath(__file__))
        olympus_datasets = []
        for dir_name in glob(f"{module_path}/dataset_*"):
            dir_name = dir_name.split("/")[-1][8:]
            olympus_datasets.append(dir_name)
        if kind not in olympus_datasets:
            message = (
                "Could not find dataset `{0}`. Please choose from one of the available "
                "datasets: {1}.".format(kind, ", ".join(list(olympus_datasets)))
            )
            Logger.log(message, "FATAL")
        # --------------------------------------------------------------
        # we will discard these arguments, so check if they are provided
        # --------------------------------------------------------------
        if data is not None:
            message = (
                "One of the Olympus datasets has been loaded via the argument `kind`, argument `data` "
                "will be discarded"
            )
            Logger.log(message, "WARNING")
        if columns is not None:
            message = (
                "One of the Olympus datasets has been loaded via the argument `kind`, argument `columns` "
                "will be discarded"
            )
            Logger.log(message, "WARNING")
        if target_names is not None:
            message = (
                "One of the Olympus datasets has been loaded via the argument `kind`, argument "
                "`target_names` will be discarded"
            )
            Logger.log(message, "WARNING")



#!/usr/bin/env python

import traceback
from olympus import Logger

try:
    import silence_tensorflow

    silence_tensorflow.silence_tensorflow()
    import tensorflow
except ModuleNotFoundError:
    # model_nn throws ModuleNotFoundError exceptions for any module that
    # it depends on but that is not installed.
    # We catch the exception, determine the missing module and pass
    # it on to the user.
    error = traceback.format_exc()
    for line in error.split("\n"):
        if "ModuleNotFoundError" in line:
            module = line.strip().strip("'").split("'")[-1]

    message = """model_nn requires {module}, which could not be found.
	Please install {module}""".format(
        module=module
    )
    Logger.log(message, "FATAL")

# ========================================================================

from .wrapper_tensorflow_model import WrapperTensorflowModel



#!/usr/bin/env python

from olympus.planners.wrapper_scipy import WrapperScipy
from olympus.objects.object_config import Config


class Simplex(WrapperScipy):

    METHOD = 'Nelder-Mead'

    # defaults are copied from scipy documentation
    # --> https://docs.scipy.org/doc/scipy/reference/optimize.minimize-neldermead.html#optimize-minimize-neldermead
    def __init__(self, goal='minimize', disp=False, maxiter=None, maxfev=None, initial_simplex=None, xatol=0.0001, fatol=0.0001,
                 adaptive=False, init_guess=None, init_guess_method='random', init_guess_seed=None):
        """
        Nelder-Mead simplex algorithm. Implementation from SciPy.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            disp (bool): Set to True to print convergence messages.
            maxiter (int): Maximum allowed number of iterations. Will default
                to N*200, where N is the number of variables, if neither maxiter
                or maxfev is set. If both maxiter and maxfev are set,
                minimization will stop at the first reached.
            maxfev (int): Maximum allowed number of function evaluations. Will
                default to N*200, where N is the number of variables, if neither
                maxiter or maxfev is set. If both maxiter and maxfev are set,
                minimization will stop at the first reached.
            initial_simplex (array_like of shape (N + 1, N)): Initial simplex.
                If given, overrides x0. initial_simplex[j,:] should contain the
                coordinates of the j-th vertex of the N+1 vertices in the
                simplex, where N is the dimension.
            xatol (float, optional): Absolute error in xopt between iterations
                that is acceptable for convergence.
            fatol (number, optional): Absolute error in func(xopt) between
                iterations that is acceptable for convergence.
            adaptive (bool, optional): Adapt algorithm parameters to
                dimensionality of problem. Useful for high-dimensional
                minimization.
            init_guess (array, optional): initial guess for the optimization
            init_guess_method (str): method to construct initial guesses if init_guess is not provided.
                Choose from: random
            init_guess_seed (str): random seed for init_guess_method
        """

        WrapperScipy.__init__(**locals())



#!/usr/bin/env python

import os

# ===============================================================================

from olympus import __home__
from olympus import Logger

from .connector_server import ConnectorServer
from .connector_github import ConnectorGithub

# ===============================================================================


class ParserDownload:
    def __init__(self, subparsers):
        self.parser = subparsers.add_parser("download", help=">> help for download")
        self.group = self.parser.add_mutually_exclusive_group(required=True)
        self.group.add_argument(
            "-n", "--name", action="store",
        )
        self.group.add_argument(
            "-l", "--list", action="store_true",
        )

    @staticmethod
    def __call__(args):
        if args.list is True:
            ParserDownload._list_datasets()
        elif args.name is not None:
            ParserDownload._get_dataset(args.name)
        else:
            Logger.log("could not parse command line arguments", "ERROR")

    @staticmethod
    def _list_datasets():
        # check github first and the server only as a backup
        for Connector in [ConnectorGithub, ConnectorServer]:
            datasets = Connector().list()
            if isinstance(datasets, list):
                Logger.log(f"found datasets: {datasets}", "INFO")
                return
        Logger.log("could not retrieve list of datasets", "ERROR")

    @staticmethod
    def _get_dataset(dataset_name):

        # check if dataset already exists
        expected_files = ["dataset.zip", "data.csv", "description.txt"]
        exists = False
        for expected_file in expected_files:
            exists = exists or os.path.isfile(
                f"{__home__}/datasets/dataset_{dataset_name}/{expected_file}"
            )
        if exists:
            Logger.log("The dataset already exists", "INFO")
            return

        # download dataset, check with github first
        for Connector in [ConnectorGithub, ConnectorServer]:
            success = Connector().get_dataset(dataset_name)
            if success:
                break


# ===============================================================================



#!/usr/bin/env python

import json
import pandas as pd

from olympus.analyzer import Analyzer

#===============================================================================

class Summarizer():
	def __init__(self, campaigns=[], include_stats = True):
		''' class to summarize the results of an Olympus benchmark into digestible
		format --> latex table, csv, xlsx

		Args:
			campaigns (list): list of Olympus campaigns
			inc_analysis (bool): specifies whether or not to include statistics in the tables

		 '''
		self.campaigns = campaigns
		self.include_stats = include_stats
		if self.include_stats:
			self.analyzer = Analyzer(campaigns = campaigns)
			self.stats_map = {'mean':   self.analyzer.get_best_mean,
							  'median': self.analyzer.get_best_median,
							  'std':    self.analyzer.get_best_std,
							  'min':    self.analyzer.get_best_min,
							  'max':    self.analyzer.get_best_max
						}


	def _load_json(self):
		''' load json file containing the saved citation information '''
		self.dataset_bib = json.load(open('olympus/summarizer/dataset_bib.json', 'r'), strict = False)
		self.planner_bib  = json.load(open('olympus/summarizer/planner_bib.json', 'r'), strict = False)

	def _sort_campaigns(self, campaigns):
		''' convert a list of campaigns to a dictionary where the keys
		are the planner names and the values are sublists of campaigns all having the
		same planner.

		Args:
			campaigns (list): a list of olympus campaign objects
		Return:
			sorted_campaigns (dict): dictionary of campaign lists
		'''
		sorted_campaigns = dict()
		for campaign in campaigns:
			if not campaign.planner_kind in sorted_campaigns.keys():
				sorted_campaigns[campaign.planner_kind] = [campaign]
			else:
				sorted_campaigns[campaign.planner_kind].append(campaign)
		return sorted_campaigns

	def _get_stats(self, sorted_campaigns, stats_list, locs, precision = 1):
		''' generate statistics for the various campigns
			possible statistics:
				DEFAULT: mean
				OPTIONAL: median, std, min, max

		Args:
			sorted_campaigns (dict): dictionary of campaign lists
			stats_lists (list): list of the statistics to compute
			locs (array-like): array of locations at which to compute the stats
			precision (int): the number of decimals to include in the table
		Returns:
			stats_dict (dict): dictionary with planner stats in it
		'''
		stats_dict = dict()
		for planner_name, campaigns in sorted_campaigns.items():
			stats = dict()
			for stat in stats_list:
				if stat == 'mean':
					res = self.stats_map[stat](campaigns = campaigns,
												locs      = locs,
												ci_method = None,
												ci_size   = 100)
				else:
					res = self.stats_map[stat](campaigns = campaigns,
									   		   locs      = locs)
				if len(res) == 1:
					res = res[0]
					res = round(res, precision)
				stats[stat] = res
			stats_dict[planner_name] = stats
		return stats_dict

	def _write_bib(self, campaigns, filename  = 'summary.bib'):
		''' write all of the dataset and planner citations to a .bib file

		Args:
			campaigns (list): a list of campiagns from which the Bibtex entries are
			referenced
		Returns:
			None
		'''
		datasets = list(set(campaign['dataset_kind'] for campaign in campaigns))
		planners = list(set(campaign['planner_kind'] for campaign in campaigns))
		with open(filename, 'w') as f:
			f.write('%==== DATASETS ======================\n\n')
			for dataset in datasets:
				f.write(self.dataset_bib[dataset]['bibtex'] + '\n')
			f.write('\n%==== PLANNERS ======================\n\n')
			for planner in planners:
				f.write(self.planner_bib[planner]['bibtex'] + '\n')
			f.close()

	def _write_table_capton(self, dataset, model_kind, goal):
		'''
		Write a caption for the latex table

		Args:
			dataset (str): the name of the dataset in the campaign
			model_kind (str): the name of the model used in the emulator
			goal (str): the name of the optimization goal (maximize or minimize)
		Returns:
			caption (str): a string table caption
		'''
		caption = f'Benchmark results on the {dataset} dataset emulated using a {model_kind} with the goal {goal}.'
		return caption

	def to_latex(self, filename = 'summary.tex', stats_list = ['mean']):
		''' Write the summary to a LaTeX table, and write the citations
		to a .bib file

		Args:
			filename (str): the name of the .tex and .bib files you wish to save
			stats_list (list): list of statistics to include in the table

		Returns:
			None
		'''
		self._load_json()
		caption = self._write_table_capton(self.campaigns[0]['dataset_kind'],
										   self.campaigns[0]['model_kind'],
										   self.campaigns[0]['goal'])

		if self.include_stats:
			# compute the stats by default, the stats are computed at the final evaluation of the campaign
			sorted_campaigns = self._sort_campaigns(self.campaigns)
			stats_dict = self._get_stats(sorted_campaigns,
										 stats_list,
										 [self.campaigns[0].observations.get_values().shape[0]])

			df = pd.DataFrame(stats_dict)
			# replace the row names with the name and the citation
			columns = df.columns
			# rename_map = dict()
			# for column in columns:
			# 	rename_map[column] = column + ' /cite{' + ','.join(self.planner_bib[column]['cite_keys']) + '}'
			# df = df.rename(columns = rename_map)
			# switch the rows and columns
			df = df.transpose()
			latex_tab = df.to_latex(index   = True,
									caption = caption,
									label   = 'tab:benchmark_results')
		else:
			planners = list(set([c.planner_kind for c in self.campaigns]))
			df = pd.DataFrame({'planner': planners})
			latex_tab = df.to_latex(index   = False,
									caption = caption,
									label   = 'tab:planners')
			# write the table
			with open(filename, 'w') as f:
				f.write(latex_tab)
				f.close()

			self._write_bib(self.campaigns)


	def to_csv(self, filename = 'summary.csv', stats_list = ['mean']):
		'''Write the summary to a .csv file

		Args:
			filename (str): the name of the .tex and .bib files you wish to save
			stats_list (list): list of statistics to include in the table

		Returns:
			None
		'''
		# TODO: What will we do with the citations here: omit for now
		if self.include_stats:
			# compute the stats
			sorted_campaigns = self._sort_campaigns(self.campaigns)
			# by default, the stats are computed at the final evaluation of the campaign
			stats_dict = self._get_stats(sorted_campaigns,
										 stats_list,
										 [self.campaigns[0].observations.get_values().shape[0]])
			df = pd.DataFrame(stats_dict).transpose()
		else:
			planners = list(set([c.planner_kind for c in self.campaigns]))
			df = pd.DataFrame({'planner': planners})
		df.to_csv(filename)



	def to_excel(self, filename = 'summary.xlsx', stats_list = ['mean']):
		'''Write the summary to an .xlsx file

		Args:
			filename (str): the name of the .tex and .bib files you wish to save
			stats_list (list): list of statistics to include in the table

		Returns:
			None
		'''
		# TODO: What will we do with the citations here: omit for now
		if self.include_stats:
			# compute the stats by default, the stats are computed at the final evaluation of the campaign
			sorted_campaigns = self._sort_campaigns(self.campaigns)
			stats_dict = self._get_stats(sorted_campaigns,
										 stats_list,
										 [self.campaigns[0].observations.get_values().shape[0]])
			df = pd.DataFrame(stats_dict).transpose()
		else:
			planners = list(set([c.planner_kind for c in self.campaigns]))
			df = pd.DataFrame({'planner': planners})
		df.to_excel(filename)



#!/usr/bin/env python

from olympus import Logger

from . import get_surfaces_list
from . import import_surface
from . import AbstractSurface


def Surface(kind='Dejong', param_dim=2):
    """Convenience function to access surfaces via a slightly higher level interface. It returns a certain surface
    with defaults arguments by keyword.

    Args:
        kind (str or AbstractPlanner): Keyword identifying one of the algorithms available in Olympus. Alternatively,
            you can pass a custom algorithm that is a subclass of AbstractPlanner.
        param_dim (int):

    Returns:
        Surface: An instance of the chosen surface.
    """
    _validate_surface_kind(kind)
    # if a string is passed, then load the corresponding wrapper
    if type(kind) == str:
        surface = import_surface(kind)
        if kind in ['Branin', 'Denali', 'Everest', 'K2', 'Kilimanjaro', 'Matterhorn', 'MontBlanc']:
            surface = surface()
            if param_dim != 2:
                message = f'Surface {kind} is only defined in 2 dimensions: setting `param_dim`=2'
                Logger.log(message, 'WARNING')
        else:
            surface = surface(param_dim=param_dim)
    # if an instance of a planner is passed, simply return the same instance
    elif isinstance(kind, AbstractSurface):
        surface = kind
    # if a custom class is passed, then that is the 'wrapper'
    elif issubclass(kind, AbstractSurface):
        surface = kind()

    return surface


def _validate_surface_kind(kind):
    # if we received a string
    if type(kind) == str:
        avail_surfaces = get_surfaces_list()
        if kind not in avail_surfaces:
            message = ('Surface "{0}" not available in Olympus. Please choose '
                       'from one of the available surfaces: {1}'.format(kind, ', '.join(avail_surfaces)))
            Logger.log(message, 'FATAL')

    # if we get an instance of a planner class
    elif isinstance(kind, AbstractSurface):
        # make sure it has the necessary methods
        for method in ['_run']:
            implementation = getattr(kind, method, None)
            if not callable(implementation):
                message = f'The object {kind} does not implement the necessary method "{method}"'
                Logger.log(message, 'FATAL')

    # if we received a custom surface class
    elif issubclass(kind, AbstractSurface):
        # make sure it has the necessary methods
        for method in ['_run']:
            implementation = getattr(kind, method, None)
            if not callable(implementation):
                message = f'The object {kind} does not implement the necessary method "{method}"'
                Logger.log(message, 'FATAL')

    # if we do not know what was passed raise an error
    else:
        message = 'Could not initialize Surface: the argument "kind" is neither a string or AbstractSurface subclass'
        Logger.log(message, 'FATAL')



#!/usr/bin/env python

from olympus.utils.misc import check_planner_module

# ----------------
# Import planner
# ----------------

planner = 'Genetic'
module  = 'deap'
link    = 'https://deap.readthedocs.io/en/master/'
check_planner_module(planner, module, link)

# ------------------------
# Import Planner Wrapper
# ------------------------
from .wrapper_deap import Genetic



#!/usr/bin/env python

import abc
import argparse 

#===============================================================================

class ActionOlympus(argparse.Action):

	def __init__(self, option_strings, dest, nargs=None, **kwargs):
		if nargs is not None:
			raise ValueError('nargs not allowed')
		argparse.Action.__init__(self, option_strings, dest, **kwargs)

	def __call__(self, parser, name_space, values, option_string=None):
		setattr(name_space, self.dest, values)
		print('VALUES', values)
		self._action(values)

	@abc.abstractmethod
	def _action(self, *args):
		pass				

#===============================================================================



#!/usr/bin/env python

#======================================================================

import traceback

from olympus import Logger

#======================================================================

try:
	import smac
except ModuleNotFoundError:

	# GPyOpt throws ModuleNotFoundError exceptions for any module that
	# it depends on but that is not installed.
	# We catch the exception, determine the missing module and pass
	# it on to the user.
	error = traceback.format_exc()
	for line in error.split('\n'):
		if 'ModuleNotFoundError' in line:
			module = line.strip().strip("'").split("'")[-1]

	message = '''SMAC requires {module}, which could not be found.
	Please install {module} and check out
	https://www.automl.org/automated-algorithm-design/algorithm-configuration/smac/ for further instructions'''.format(module = module)
	Logger.log(message, 'FATAL')

#======================================================================

from .wrapper_smac import Smac



#!/usr/bin/env python

# ===============================================================================

from .campaigns import Campaign
from .databases import Database
from .emulators import Emulator
from .evaluators import Evaluator
from .objects import Object
from .plotter import Plotter
from .planners import Planner, get_planners_list
from .surfaces import Surface

from . import __home__, __scratch__

# ===============================================================================


class Olympus(Object):

    """ Master class of the olympus package
    """

    def __init__(self, *args, **kwargs):
        Object.__init__(**locals())
        self.home = __home__
        self.scratch = __scratch__
        self.database = Database()

    # *** Production ****************************************************************
    def run(
        self,
        planner="Phoenics",
        dataset="alkox",
        model="BayesNeuralNet",
        goal="default",
        campaign=Campaign(),
        database=Database(),
        num_iter=3,
    ):

        emulator = Emulator(dataset=dataset, model=model)
        if goal == "default":
            goal = emulator.goal
        planner_ = Planner(kind=planner, goal=goal)

        # set links
        self.planner = planner_
        self.emulator = emulator
        self.campaign = campaign
        self.database = database

        # define evaluator and optimize
        self.evaluator = Evaluator(
            planner=planner_, emulator=emulator, campaign=campaign, database=database,
        )
        self.evaluator.optimize(num_iter=num_iter)

    def run_analytic(
        self,
        planner="Phoenics",
        surface="dejong",
        param_dim=2,
        goal="minimize",
        campaign=Campaign(),
        database=Database(),
        num_iter=3,
    ):

        self.planner = Planner(kind=planner, goal=goal)
        self.surface = Surface(name=surface, param_dim=param_dim)
        self.campaign = campaign
        self.database = database
        self.evaluator = Evaluator(
            planner=self.planner,
            emulator=self.surface,
            campaign=self.campaign,
            database=self.database,
        )
        self.evaluator.optimize(num_iter=num_iter)

    def benchmark(
        self,
        dataset="alkox",
        planners="all",
        database=Database(kind="sqlite"),
        num_ind_runs=5,
        num_iter=3
    ):
        """
        Args:
            dataset (str): the dataset to use
        """
        if planners == "all":
            planners = get_planners_list()

        for planner in planners:
            for _ in range(num_ind_runs):
                self.run(
                    planner=planner,
                    dataset=dataset,
                    database=database,
                    campaign=Campaign(),
                    num_iter=num_iter
                )


# *** Analysis ******************************************************************

# 	def load_database(self, file_name):
# 		''' connects to a database previously written by olympus
#
# 		Args:
# 			file_name (str or list): path and name of the database file
# 		'''
# 		if hasattr(self, 'database'):
# 			self.database.from_file(file_name)
# 		else:
# 			self.database = Database().from_file(file_name)
#
#
# 	def get_campaigns(self, file_names=[]):
# 		if len(file_names) > 0:
# 			return Database().from_file(file_names).get_campaigns()
# 		else:
# 			return self.database.get_campaigns()

# ===============================================================================



#!/usr/bin/env python

#==============================================================================

import sqlalchemy

#==============================================================================

from olympus.campaigns import Campaign
from olympus.databases import AbstractDatabase
from olympus.databases.database_sqlite import SqliteInterface

#==============================================================================

class Wrapper_sqlite(AbstractDatabase):

	''' implements a simple sqlite database

	Architecture of the database:
		image of the `Campaign` class, where entries encode individual objects
	'''

	ATT_KIND      = {'type': 'string', 'default': 'sqlite'}
	ATT_FILE_TYPE = {'type': 'string', 'default': 'sqlite', 'valid': ['db', 'sqlite']}

	def __init__(self, *args, **kwargs):
		AbstractDatabase.__init__(self, *args, **kwargs)
		self.logged_campaign_ids = []

	def __getattr__(self, prop):
		if prop == 'db':
			self.db = SqliteInterface(self.name, self.file_name, self.db_attrs())
			return self.db
		else:
			return AbstractDatabase.__getattr__(self, prop)


	def db_attrs(self):
		db_attrs = Campaign().defaults
		return db_attrs

	def _get_campaigns(self):
		campaigns = self.db.fetch_all()
		return campaigns

	def _list_campaign_ids(self):
		campaigns = self.db.fetch_all()
		return [campaign.id for campaign in campaigns]

	def _load_db(self):
		return self.db.fetch_all()

	def _log_campaign(self, campaign):
		if campaign.id in self.logged_campaign_ids:
			self.db.update_all({'id': campaign.id}, campaign.to_dict())
		else:
			self.db.add(campaign.to_dict())
			self.logged_campaign_ids.append(campaign.id)



#!/usr/bin/env python

# ===============================================================================

import os

from olympus import __scratch__
from olympus import Logger
from olympus.objects import Object
from olympus.utils import generate_id

# ===============================================================================


class AbstractDatabase(Object):

    """ bridge to specific databases; implements in-memory cache
    """

    ATT_KIND = {"type": "string", "default": "abstract"}
    ATT_NAME = {"type": "string", "default": lambda: "olympus_{}".format(generate_id())}
    ATT_PATH = {"type": "string", "default": __scratch__}

    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)

    def __repr__(self):
        return f"<Database (name={self.name}, kind={self.kind})>"

    @property
    def abstract_methods(self):
        return self.ABSTRACT_METHODS

    @property
    def is_processing(self):
        return self.db.is_processing

    @property
    def db_exists(self):
        return os.path.isfile(self.file_name)

    @property
    def file_name(self):
        return os.path.join(self.path, "{}.{}".format(self.name, self.file_type))

    @property
    def file_types(self):
        return self.ATT_FILE_TYPE["valid"]

    def _from_file_name(self, file_name):
        self.path = "/".join(file_name.split("/")[:-1])
        self.name = file_name.split("/")[-1].split(".")[0]
        self.file_type = file_name.split("/")[-1].split(".")[1]

    def from_file(self, file_name):
        self._from_file_name(file_name)
        if not self.db_exists:
            Logger.log("Could not find database file {}".format(file_name), "ERROR")
            return None
        self._load_db()

    def get_campaigns(self):
        return self._get_campaigns()

    def list_campaigns(self):
        return self._list_campaign_ids()

    def log_campaign(self, campaign):
        self._log_campaign(campaign)

    ABSTRACT_METHODS = locals()


# ===============================================================================



#!/usr/bin/env python

#===============================================================================

import time
import sqlalchemy as sql

from olympus.utils import thread

#===============================================================================

class AddEntry:

    def __init__(self, db, table, entry):
        self.db    = db
        self.table = table
        self.entry = entry

    def execute(self):
        with self.db.connect() as conn:
            conn.execute(self.table.insert(), self.entry)
            conn.close()

#===============================================================================

class FetchEntries:

    def __init__(self, db, table, select, name = 'name'):
        self.db     = db
        self.table  = table
        self.select = select
        self.name   = name

        self.entries         = None
        self.executed        = False
        self.entries_fetched = False

    def execute(self):
        with self.db.connect() as conn:
            selected = conn.execute(self.select)
            entries  = selected.fetchall()
            conn.close()
        self.entries  = entries
        self.executed = True

    def get_entries(self):
        iter_index = 0
        while not self.executed:
            time.sleep(0.1)
        self.entries_fetched = True
        return self.entries

#===============================================================================

class UpdateEntries:

    def __init__(self, db, table, updates):
        self.db      = db
        self.table   = table
        self.updates = updates

    def execute(self):
        with self.db.connect() as conn:
            updated = conn.execute(self.updates)
            conn.close()



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface


class StyblinskiTang(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """Styblinski-Tang function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        x = (-2.903534 + 5) / 10  # rescale onto unit square
        params = [x] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        return None

    def _run(self, params):
        params = np.array(params)
        params = 10 * params - 5  # rescale onto [-5, 5]
        result = 0.5 * np.sum(params**4 - 16*params**2 + 5*params)

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

from olympus.utils.misc import check_module
from olympus.surfaces import get_surfaces_list

surface = 'GaussianMixture'
module  = 'scipy'
link    = 'https://docs.scipy.org/doc/scipy/reference/'

message = '''Surface {surface} requires {module} which could not be found. Please install
{module} and check out {link} for further instructions or use another
available planner: {planners_list}'''
check_module(module, message, planner=surface, planners_list=', '.join(get_surfaces_list()), link=link)

from .wrapper_gaussian_mixture import GaussianMixture



#!/usr/bin/env python

import time

from olympus.objects.object_config     import Config
from olympus.objects                   import ParameterVector
from olympus.planners.abstract_planner import AbstractPlanner
from olympus.planners.utils_planner    import get_init_guess, get_bounds
from olympus.utils                     import daemon


#===============================================================================

class DifferentialEvolution(AbstractPlanner):

    # defaults are copied from scipy documentation
    # --> https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html
    def __init__(self, goal='minimize', args=(), strategy='best1bin', maxiter=1000, popsize=15,
                 tol=0.01, mutation=(0.5, 1), recombination=0.7, seed=None,
                 callback=None, disp=False, polish=True, init='latinhypercube',
                 atol=0, updating='immediate', workers=1,
                 init_guess=None, init_guess_method='random', init_guess_seed=None):
        """Finds the global minimum of a multivariate function.
            Differential Evolution is stochastic in nature (does not use gradient
            methods) to find the minimium, and can search large areas of candidate
            space, but often requires larger numbers of function evaluations than
            conventional gradient based techniques.
            The algorithm is due to Storn and Price [1]_.

            Parameters
            ----------
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            func : callable
                The objective function to be minimized.  Must be in the form
                ``f(x, *args)``, where ``x`` is the argument in the form of a 1-D array
                and ``args`` is a  tuple of any additional fixed parameters needed to
                completely specify the function.
            bounds : sequence
                Bounds for variables.  ``(min, max)`` pairs for each element in ``x``,
                defining the lower and upper bounds for the optimizing argument of
                `func`. It is required to have ``len(bounds) == len(x)``.
                ``len(bounds)`` is used to determine the number of parameters in ``x``.
            args : tuple, optional
                Any additional fixed parameters needed to
                completely specify the objective function.
            strategy : str, optional
                The differential evolution strategy to use. Should be one of:
                    - 'best1bin'
                    - 'best1exp'
                    - 'rand1exp'
                    - 'randtobest1exp'
                    - 'currenttobest1exp'
                    - 'best2exp'
                    - 'rand2exp'
                    - 'randtobest1bin'
                    - 'currenttobest1bin'
                    - 'best2bin'
                    - 'rand2bin'
                    - 'rand1bin'
                The default is 'best1bin'.
            maxiter : int, optional
                The maximum number of generations over which the entire population is
                evolved. The maximum number of function evaluations (with no polishing)
                is: ``(maxiter + 1) * popsize * len(x)``
            popsize : int, optional
                A multiplier for setting the total population size.  The population has
                ``popsize * len(x)`` individuals (unless the initial population is
                supplied via the `init` keyword).
            tol : float, optional
                Relative tolerance for convergence, the solving stops when
                ``np.std(pop) <= atol + tol * np.abs(np.mean(population_energies))``,
                where and `atol` and `tol` are the absolute and relative tolerance
                respectively.
            mutation : float or tuple(float, float), optional
                The mutation constant. In the literature this is also known as
                differential weight, being denoted by F.
                If specified as a float it should be in the range [0, 2].
                If specified as a tuple ``(min, max)`` dithering is employed. Dithering
                randomly changes the mutation constant on a generation by generation
                basis. The mutation constant for that generation is taken from
                ``U[min, max)``. Dithering can help speed convergence significantly.
                Increasing the mutation constant increases the search radius, but will
                slow down convergence.
            recombination : float, optional
                The recombination constant, should be in the range [0, 1]. In the
                literature this is also known as the crossover probability, being
                denoted by CR. Increasing this value allows a larger number of mutants
                to progress into the next generation, but at the risk of population
                stability.
            seed : int or `np.random.RandomState`, optional
                If `seed` is not specified the `np.RandomState` singleton is used.
                If `seed` is an int, a new `np.random.RandomState` instance is used,
                seeded with seed.
                If `seed` is already a `np.random.RandomState instance`, then that
                `np.random.RandomState` instance is used.
                Specify `seed` for repeatable minimizations.
            disp : bool, optional
                Display status messages
            callback : callable, `callback(xk, convergence=val)`, optional
                A function to follow the progress of the minimization. ``xk`` is
                the current value of ``x0``. ``val`` represents the fractional
                value of the population convergence.  When ``val`` is greater than one
                the function halts. If callback returns `True`, then the minimization
                is halted (any polishing is still carried out).
            polish : bool, optional
                If True (default), then `scipy.optimize.minimize` with the `L-BFGS-B`
                method is used to polish the best population member at the end, which
                can improve the minimization slightly.
            init : str or array-like, optional
                Specify which type of population initialization is performed. Should be
                one of:
                    - 'latinhypercube'
                    - 'random'
                    - array specifying the initial population. The array should have
                      shape ``(M, len(x))``, where len(x) is the number of parameters.
                      `init` is clipped to `bounds` before use.
                The default is 'latinhypercube'. Latin Hypercube sampling tries to
                maximize coverage of the available parameter space. 'random'
                initializes the population randomly - this has the drawback that
                clustering can occur, preventing the whole of parameter space being
                covered. Use of an array to specify a population subset could be used,
                for example, to create a tight bunch of initial guesses in an location
                where the solution is known to exist, thereby reducing time for
                convergence.
            atol : float, optional
                Absolute tolerance for convergence, the solving stops when
                ``np.std(pop) <= atol + tol * np.abs(np.mean(population_energies))``,
                where and `atol` and `tol` are the absolute and relative tolerance
                respectively.
            updating : {'immediate', 'deferred'}, optional
                If ``'immediate'``, the best solution vector is continuously updated
                within a single generation [4]_. This can lead to faster convergence as
                trial vectors can take advantage of continuous improvements in the best
                solution.
                With ``'deferred'``, the best solution vector is updated once per
                generation. Only ``'deferred'`` is compatible with parallelization, and
                the `workers` keyword can over-ride this option.
                .. versionadded:: 1.2.0
            workers : int or map-like callable, optional
                If `workers` is an int the population is subdivided into `workers`
                sections and evaluated in parallel (uses `multiprocessing.Pool`).
                Supply -1 to use all available CPU cores.
                Alternatively supply a map-like callable, such as
                `multiprocessing.Pool.map` for evaluating the population in parallel.
                This evaluation is carried out as ``workers(func, iterable)``.
                This option will override the `updating` keyword to
                ``updating='deferred'`` if ``workers != 1``.
                Requires that `func` be pickleable.
                .. versionadded:: 1.2.0
            init_guess (array, optional): initial guess for the optimization
            init_guess_method (str): method to construct initial guesses if init_guess is not provided.
                Choose from: random
            init_guess_seed (str): random seed for init_guess_method
            Returns
            -------
            res : OptimizeResult
                The optimization result represented as a `OptimizeResult` object.
                Important attributes are: ``x`` the solution array, ``success`` a
                Boolean flag indicating if the optimizer exited successfully and
                ``message`` which describes the cause of the termination. See
                `OptimizeResult` for a description of other attributes.  If `polish`
                was employed, and a lower minimum was obtained by the polishing, then
                OptimizeResult also contains the ``jac`` attribute.
            Notes
            -----
            Differential evolution is a stochastic population based method that is
            useful for global optimization problems. At each pass through the population
            the algorithm mutates each candidate solution by mixing with other candidate
            solutions to create a trial candidate. There are several strategies [2]_ for
            creating trial candidates, which suit some problems more than others. The
            'best1bin' strategy is a good starting point for many systems. In this
            strategy two members of the population are randomly chosen. Their difference
            is used to mutate the best member (the `best` in `best1bin`), :math:`b_0`,
            so far:
            .. math::
                b' = b_0 + mutation * (population[rand0] - population[rand1])
            A trial vector is then constructed. Starting with a randomly chosen 'i'th
            parameter the trial is sequentially filled (in modulo) with parameters from
            ``b'`` or the original candidate. The choice of whether to use ``b'`` or the
            original candidate is made with a binomial distribution (the 'bin' in
            'best1bin') - a random number in [0, 1) is generated.  If this number is
            less than the `recombination` constant then the parameter is loaded from
            ``b'``, otherwise it is loaded from the original candidate.  The final
            parameter is always loaded from ``b'``.  Once the trial candidate is built
            its fitness is assessed. If the trial is better than the original candidate
            then it takes its place. If it is also better than the best overall
            candidate it also replaces that.
            To improve your chances of finding a global minimum use higher `popsize`
            values, with higher `mutation` and (dithering), but lower `recombination`
            values. This has the effect of widening the search radius, but slowing
            convergence.
            By default the best solution vector is updated continuously within a single
            iteration (``updating='immediate'``). This is a modification [4]_ of the
            original differential evolution algorithm which can lead to faster
            convergence as trial vectors can immediately benefit from improved
            solutions. To use the original Storn and Price behaviour, updating the best
            solution once per iteration, set ``updating='deferred'``.
            .. versionadded:: 0.15.0
            Examples
            --------
            Let us consider the problem of minimizing the Rosenbrock function. This
            function is implemented in `rosen` in `scipy.optimize`.
            >>> from scipy.optimize import rosen, differential_evolution
            >>> bounds = [(0,2), (0, 2), (0, 2), (0, 2), (0, 2)]
            >>> result = differential_evolution(rosen, bounds)
            >>> result.x, result.fun
            (array([1., 1., 1., 1., 1.]), 1.9216496320061384e-19)
            Now repeat, but with parallelization.
            >>> bounds = [(0,2), (0, 2), (0, 2), (0, 2), (0, 2)]
            >>> result = differential_evolution(rosen, bounds, updating='deferred',
            ...                                 workers=2)
            >>> result.x, result.fun
            (array([1., 1., 1., 1., 1.]), 1.9216496320061384e-19)
            Next find the minimum of the Ackley function
            (https://en.wikipedia.org/wiki/Test_functions_for_optimization).
            >>> from scipy.optimize import differential_evolution
            >>> import numpy as np
            >>> def ackley(x):
            ...     arg1 = -0.2 * np.sqrt(0.5 * (x[0] ** 2 + x[1] ** 2))
            ...     arg2 = 0.5 * (np.cos(2. * np.pi * x[0]) + np.cos(2. * np.pi * x[1]))
            ...     return -20. * np.exp(arg1) - np.exp(arg2) + 20. + np.e
            >>> bounds = [(-5, 5), (-5, 5)]
            >>> result = differential_evolution(ackley, bounds)
            >>> result.x, result.fun
            (array([ 0.,  0.]), 4.4408920985006262e-16)
            References
            ----------
            .. [1] Storn, R and Price, K, Differential Evolution - a Simple and
                   Efficient Heuristic for Global Optimization over Continuous Spaces,
                   Journal of Global Optimization, 1997, 11, 341 - 359.
            .. [2] http://www1.icsi.berkeley.edu/~storn/code.html
            .. [3] http://en.wikipedia.org/wiki/Differential_evolution
            .. [4] Wormington, M., Panaccione, C., Matney, K. M., Bowen, D. K., -
                   Characterization of structures from X-ray scattering data using
                   genetic algorithms, Phil. Trans. R. Soc. Lond. A, 1999, 357,
                   2827-2848
            """
        kwargs = locals().copy()
        del kwargs['self']

        self.goal = kwargs['goal']
        for attr in ['goal', 'init_guess', 'init_guess_method', 'init_guess_seed']:
            del kwargs[attr]

        self.kwargs = kwargs
        self.has_minimizer = False
        self.is_converged  = False
        AbstractPlanner.__init__(**locals())

    def _set_param_space(self, param_space):
        self.param_space = param_space
        self.bounds      = get_bounds(param_space)
        if self.init_guess is None:
            self.init_guess = get_init_guess(param_space, method=self.init_guess_method, random_seed=self.init_guess_seed)

    def _tell(self, observations):
        self._params = observations.get_params(as_array=False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
        if len(self._values) > 0:
            self.RECEIVED_VALUES.append(self._values[-1])

    def _priv_evaluator(self, params):
        params = self._project_into_domain(params)
        self.SUBMITTED_PARAMS.append(params)
        while len(self.RECEIVED_VALUES) == 0:
            time.sleep(0.1)
        value = self.RECEIVED_VALUES.pop(0)
        return value

    @daemon
    def create_optimizer(self):
        from scipy.optimize import differential_evolution
        _ = differential_evolution(self._priv_evaluator, bounds=self.bounds, **self.kwargs)
        self.is_converged = True

    def _ask(self):
        if self.has_minimizer is False:
            self.create_optimizer()
            self.has_minimizer = True
        while len(self.SUBMITTED_PARAMS) == 0:
            time.sleep(0.1)
            if self.is_converged:
                return ParameterVector().from_dict(self._params[-1])
        params = self.SUBMITTED_PARAMS.pop(0)
        return ParameterVector().from_array(params, self.param_space)
